from pathlib import Path
from typing import cast
from unittest import mock

from dagster import (
    AssetKey,
    AssetsDefinition,
    AssetSpec,
    Definitions,
    asset,
    asset_check,
    executor,
    job,
    logger,
    multi_asset,
    schedule,
    sensor,
)
from dagster._core.code_pointer import CodePointer
from dagster._core.definitions.assets.definition.asset_dep import AssetDep
from dagster._core.definitions.job_definition import JobDefinition
from dagster._core.definitions.reconstruct import initialize_repository_def_from_pointer
from dagster._utils.test.definitions import (
    definitions,
    scoped_reconstruction_metadata,
    unwrap_reconstruction_metadata,
)
from dagster_airlift.constants import TASK_MAPPING_METADATA_KEY
from dagster_airlift.core import assets_with_task_mappings, build_defs_from_airflow_instance
from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.load_defs import (
    build_job_based_airflow_defs,
    enrich_airflow_mapped_assets,
    load_airflow_dag_asset_specs,
)
from dagster_airlift.core.multiple_tasks import assets_with_multiple_task_mappings
from dagster_airlift.core.serialization.compute import (
    build_airlift_metadata_mapping_info,
    compute_serialized_data,
)
from dagster_airlift.core.serialization.defs_construction import make_default_dag_asset_key
from dagster_airlift.core.serialization.serialized_data import (
    DagHandle,
    SerializedAirflowDefinitionsData,
    TaskHandle,
)
from dagster_airlift.core.utils import is_task_mapped_asset_spec, metadata_for_task_mapping
from dagster_airlift.test import asset_spec, make_instance
from dagster_shared.serdes import deserialize_value
from dagster_test.utils.definitions_execute_in_process import get_job_from_defs

from dagster_airlift_tests.unit_tests.conftest import (
    assert_dependency_structure_in_assets,
    fully_loaded_repo_from_airflow_asset_graph,
    load_definitions_airflow_asset_graph,
)


@executor  # pyright: ignore[reportCallIssue,reportArgumentType]
def nonstandard_executor(init_context):
    pass


@logger  # pyright: ignore[reportCallIssue,reportArgumentType]
def nonstandard_logger(init_context):
    pass


@sensor(job_name="the_job")
def some_sensor():
    pass


@schedule(cron_schedule="0 0 * * *", job_name="the_job")
def some_schedule():
    pass


@asset
def a():
    pass


b_spec = AssetSpec(key="b")


@asset_check(asset=a)  # pyright: ignore[reportArgumentType]
def a_check():
    pass


@job
def the_job():
    pass


def make_test_dag_asset_key(dag_id: str) -> AssetKey:
    return make_default_dag_asset_key("test_instance", dag_id)


def test_defs_passthrough() -> None:
    """Test that passed-through definitions are present in the final definitions."""
    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance({"dag": ["task"]}),
        defs=Definitions(
            assets=[a, b_spec],
            asset_checks=[a_check],
            jobs=[the_job],
            sensors=[some_sensor],
            schedules=[some_schedule],
            loggers={"the_logger": nonstandard_logger},  # pyright: ignore[reportArgumentType]
            executor=nonstandard_executor,  # pyright: ignore[reportArgumentType]
        ),
    )
    assert defs.executor == nonstandard_executor
    assert defs.loggers
    assert len(defs.loggers) == 1
    assert next(iter(defs.loggers.keys())) == "the_logger"
    assert defs.sensors
    assert len(list(defs.sensors)) == 2
    our_sensor = next(
        iter(sensor_def for sensor_def in defs.sensors if sensor_def.name == "some_sensor")
    )
    assert our_sensor == some_sensor
    assert defs.schedules
    assert len(list(defs.schedules)) == 1
    assert next(iter(defs.schedules)) == some_schedule
    assert defs.jobs
    assert len(list(defs.jobs)) == 1
    assert next(iter(defs.jobs)) == the_job
    repo = defs.get_repository_def()
    # Ensure that asset specs get properly coerced into asset defs
    assert set(repo.assets_defs_by_key.keys()) == {
        a.key,
        b_spec.key,
        make_test_dag_asset_key("dag"),
    }
    assert isinstance(repo.assets_defs_by_key[b_spec.key], AssetsDefinition)


def test_coerce_specs() -> None:
    """Test that asset specs are properly coerced into asset keys."""
    # Initialize an airflow instance with a dag "dag", which contains a task "task". There are no task instances or runs.

    spec = AssetSpec(key="a", metadata=metadata_for_task_mapping(task_id="task", dag_id="dag"))
    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance({"dag": ["task"]}),
        defs=Definitions(
            assets=[spec],
        ),
    )
    repo = defs.get_repository_def()
    assert len(repo.assets_defs_by_key) == 2
    assert AssetKey("a") in repo.assets_defs_by_key
    assets_def = repo.assets_defs_by_key[AssetKey("a")]
    # Asset metadata properties have been glommed onto the asset
    spec = next(iter(assets_def.specs))
    assert spec.metadata["Dag ID"] == "dag"


def test_invalid_dagster_named_tasks_and_dags() -> None:
    """Test that invalid dagster names are converted to valid names."""
    a = AssetKey("a")
    spec = AssetSpec(
        key=a,
        metadata=metadata_for_task_mapping(task_id="task-with-hyphens", dag_id="dag-with-hyphens"),
    )
    airflow_instance = make_instance({"dag-with-hyphens": ["task-with-hyphens"]})
    defs = build_defs_from_airflow_instance(
        airflow_instance=airflow_instance,
        defs=Definitions(
            assets=[spec],
        ),
    )

    repo = defs.get_repository_def()
    assert len(repo.assets_defs_by_key) == 2
    assert a in repo.assets_defs_by_key
    assets_def = repo.assets_defs_by_key[a]
    assert not assets_def.is_executable

    assert make_test_dag_asset_key("dag-with-hyphens") in repo.assets_defs_by_key
    dag_def = repo.assets_defs_by_key[
        make_default_dag_asset_key(airflow_instance.name, "dag_with_hyphens")
    ]
    assert not dag_def.is_executable


def has_single_task_handle(spec: AssetSpec, dag_id: str, task_id: str):
    assert len(spec.metadata[TASK_MAPPING_METADATA_KEY]) == 1
    task_handle_dict = next(iter(spec.metadata[TASK_MAPPING_METADATA_KEY]))
    return task_handle_dict["dag_id"] == dag_id and task_handle_dict["task_id"] == task_id


def test_transitive_asset_deps() -> None:
    """Test that cross-dag transitive asset dependencies are correctly generated."""
    # Asset graph is a -> b -> c where a and c are in different dags, and b isn't in any dag.
    repo_def = fully_loaded_repo_from_airflow_asset_graph(
        assets_per_task={
            "dag1": {"task": [("a", [])]},
            "dag2": {"task": [("c", ["b"])]},
        },
        additional_defs=Definitions(assets=[AssetSpec(key="b", deps=["a"])]),
    )
    repo_def.load_all_definitions()
    airflow_instance = make_instance(dag_and_task_structure={"dag1": ["task"], "dag2": ["task"]})
    dag1_key = make_default_dag_asset_key(instance_name=airflow_instance.name, dag_id="dag1")
    dag2_key = make_default_dag_asset_key(instance_name=airflow_instance.name, dag_id="dag2")
    a_key = AssetKey(["a"])
    b_key = AssetKey(["b"])
    c_key = AssetKey(["c"])
    assert len(repo_def.assets_defs_by_key) == 5
    assert set(repo_def.assets_defs_by_key.keys()) == {
        dag1_key,
        dag2_key,
        a_key,
        b_key,
        c_key,
    }

    dag1_asset = repo_def.assets_defs_by_key[dag1_key]
    assert [dep.asset_key for dep in next(iter(dag1_asset.specs)).deps] == [a_key]

    dag2_asset = repo_def.assets_defs_by_key[dag2_key]
    assert [dep.asset_key for dep in next(iter(dag2_asset.specs)).deps] == [c_key]

    a_asset = repo_def.assets_defs_by_key[a_key]
    assert [dep.asset_key for dep in next(iter(a_asset.specs)).deps] == []
    assert has_single_task_handle(next(iter(a_asset.specs)), "dag1", "task")

    b_asset = repo_def.assets_defs_by_key[b_key]
    assert [dep.asset_key for dep in next(iter(b_asset.specs)).deps] == [a_key]
    assert not is_task_mapped_asset_spec(next(iter(b_asset.specs)))

    c_asset = repo_def.assets_defs_by_key[c_key]
    assert [dep.asset_key for dep in next(iter(c_asset.specs)).deps] == [b_key]
    assert has_single_task_handle(next(iter(c_asset.specs)), "dag2", "task")


def test_peered_dags() -> None:
    """Test peered dags show up, and that linkage is preserved downstream of dags."""
    defs = load_definitions_airflow_asset_graph(
        assets_per_task={
            "dag1": {"task": []},
            "dag2": {"task": []},
            "dag3": {"task": []},
        },
        additional_defs=Definitions(
            assets=[AssetSpec(key="a", deps=[make_test_dag_asset_key("dag1")])]
        ),
    )
    assert defs.assets
    repo_def = defs.get_repository_def()
    repo_def.load_all_definitions()
    assert len(repo_def.assets_defs_by_key) == 4
    assert_dependency_structure_in_assets(
        repo_def=repo_def,
        expected_deps={
            make_test_dag_asset_key("dag1").to_user_string(): [],
            make_test_dag_asset_key("dag2").to_user_string(): [],
            make_test_dag_asset_key("dag3").to_user_string(): [],
            "a": [make_test_dag_asset_key("dag1").to_user_string()],
        },
    )
    for dag_asset_key in [
        make_test_dag_asset_key("dag1"),
        make_test_dag_asset_key("dag2"),
        make_test_dag_asset_key("dag3"),
    ]:
        dag_asset_spec = repo_def.assets_defs_by_key[dag_asset_key].specs_by_key[dag_asset_key]
        assert "dagster/kind/airflow" in dag_asset_spec.tags


def test_observed_assets() -> None:
    """Test that observed assets are properly linked to dags."""
    # Asset graph structure:
    #   a
    #  / \
    # b   c
    #  \ /
    #   d
    #  / \
    # e   f
    defs = load_definitions_airflow_asset_graph(
        assets_per_task={
            "dag": {
                "task1": [("a", []), ("b", ["a"]), ("c", ["a"])],
                "task2": [("d", ["b", "c"]), ("e", ["d"]), ("f", ["d"])],
            },
        },
    )
    assert defs.assets
    repo_def = defs.get_repository_def()
    repo_def.load_all_definitions()
    repo_def.load_all_definitions()
    assert len(repo_def.assets_defs_by_key) == 7
    assert_dependency_structure_in_assets(
        repo_def=repo_def,
        expected_deps={
            "a": [],
            "b": ["a"],
            "c": ["a"],
            "d": ["b", "c"],
            "e": ["d"],
            "f": ["d"],
            # Only leaf assets should be immediately upstream of the dag
            make_test_dag_asset_key("dag").to_user_string(): ["e", "f"],
        },
    )
    for key_str in ["a", "b", "c", "d", "e", "f"]:
        asset_spec = repo_def.assets_defs_by_key[AssetKey(key_str)].specs_by_key[AssetKey(key_str)]
        assert "dagster/kind/airliftmapped" in asset_spec.tags


def test_local_airflow_instance() -> None:
    """Test that a local-backed airflow instance can be correctly peered, and errors when the correct info can't be found."""
    defs = load_definitions_airflow_asset_graph(
        assets_per_task={
            "dag": {"task": [("a", [])]},
        },
        create_assets_defs=True,
    )

    assert defs.assets
    repo_def = defs.get_repository_def()

    defs = load_definitions_airflow_asset_graph(
        assets_per_task={
            "dag": {"task": [("a", [])]},
        },
        create_assets_defs=True,
    )
    repo_def = defs.get_repository_def()
    assert defs.assets
    repo_def = defs.get_repository_def()
    assert len(repo_def.assets_defs_by_key) == 2


@definitions
def airflow_instance_defs() -> Definitions:
    a = AssetKey("a")
    spec = AssetSpec(
        key=a,
        metadata=metadata_for_task_mapping(task_id="task", dag_id="dag"),
    )
    instance = make_instance({"dag": ["task"]})
    passed_in_defs = Definitions(assets=[spec])

    return build_defs_from_airflow_instance(airflow_instance=instance, defs=passed_in_defs)


def test_cached_loading() -> None:
    repository_def = initialize_repository_def_from_pointer(
        CodePointer.from_python_file(str(Path(__file__)), "airflow_instance_defs", None),
    )
    assert repository_def.repository_load_data
    assert len(repository_def.repository_load_data.reconstruction_metadata) == 1
    assert (
        "dagster-airlift/source/test_instance"
        in repository_def.repository_load_data.reconstruction_metadata
    )
    assert isinstance(
        repository_def.repository_load_data.reconstruction_metadata[
            "dagster-airlift/source/test_instance"
        ].value,
        str,
    )
    assert isinstance(
        deserialize_value(
            repository_def.repository_load_data.reconstruction_metadata[
                "dagster-airlift/source/test_instance"
            ].value
        ),
        SerializedAirflowDefinitionsData,
    )

    with scoped_reconstruction_metadata(unwrap_reconstruction_metadata(repository_def)):
        with mock.patch(
            "dagster_airlift.core.serialization.compute.compute_serialized_data",
            wraps=compute_serialized_data,
        ) as mock_compute_serialized_data:
            reloaded_repo_def = initialize_repository_def_from_pointer(
                CodePointer.from_python_file(str(Path(__file__)), "airflow_instance_defs", None),
            )
            assert mock_compute_serialized_data.call_count == 0
            assert reloaded_repo_def.assets_defs_by_key
            assert len(list(reloaded_repo_def.assets_defs_by_key.keys())) == 2
            assert {
                key
                for assets_def in reloaded_repo_def.assets_defs_by_key.values()
                for key in cast("AssetsDefinition", assets_def).keys
            } == {AssetKey("a"), make_test_dag_asset_key("dag")}


def test_multiple_tasks_per_asset(init_load_context: None) -> None:
    """Test behavior for a single AssetsDefinition where different specs map to different airflow tasks/dags."""

    @multi_asset(
        specs=[
            AssetSpec(key="a", metadata=metadata_for_task_mapping(task_id="task1", dag_id="dag1")),
            AssetSpec(key="b", metadata=metadata_for_task_mapping(task_id="task2", dag_id="dag2")),
        ],
        name="multi_asset",
    )
    def my_asset():
        pass

    instance = make_instance({"dag1": ["task1"], "dag2": ["task2"]})
    defs = build_defs_from_airflow_instance(
        airflow_instance=instance,
        defs=Definitions(assets=[my_asset]),
    )
    assert defs.assets
    # 3 Full assets definitions, but 4 keys
    assert len(list(defs.assets)) == 3
    assert {
        key for assets_def in defs.assets for key in cast("AssetsDefinition", assets_def).keys
    } == {
        AssetKey("a"),
        AssetKey("b"),
        make_test_dag_asset_key("dag1"),
        make_test_dag_asset_key("dag2"),
    }
    repo_def = defs.get_repository_def()
    a_and_b_asset = repo_def.assets_defs_by_key[AssetKey("a")]
    a_spec = next(iter(spec for spec in a_and_b_asset.specs if spec.key == AssetKey("a")))
    assert has_single_task_handle(a_spec, "dag1", "task1")
    b_spec = next(iter(spec for spec in a_and_b_asset.specs if spec.key == AssetKey("b")))
    assert has_single_task_handle(b_spec, "dag2", "task2")


def test_multiple_tasks_to_single_asset_metadata() -> None:
    instance = make_instance({"dag1": ["task1"], "dag2": ["task2"]})

    @asset(
        metadata={
            TASK_MAPPING_METADATA_KEY: [
                {"dag_id": "dag1", "task_id": "task1"},
                {"dag_id": "dag2", "task_id": "task2"},
            ]
        }
    )
    def an_asset() -> None: ...

    defs = build_defs_from_airflow_instance(
        airflow_instance=instance, defs=Definitions(assets=[an_asset])
    )

    assert defs.assets
    assert len(list(defs.assets)) == 3  # two dags and one asset

    assert defs.resolve_asset_graph().assets_def_for_key(AssetKey("an_asset")).specs_by_key[
        AssetKey("an_asset")
    ].metadata[TASK_MAPPING_METADATA_KEY] == [
        {"dag_id": "dag1", "task_id": "task1"},
        {"dag_id": "dag2", "task_id": "task2"},
    ]


def test_multiple_tasks_with_multiple_task_mappings() -> None:
    @asset
    def other_asset() -> None: ...

    @asset(deps=[other_asset])
    def scheduled_twice() -> None: ...

    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            {"weekly_dag": ["task1"], "daily_dag": ["task1"], "other_dag": ["task1"]}
        ),
        defs=Definitions(
            assets=[
                other_asset,
                *assets_with_multiple_task_mappings(
                    assets=[scheduled_twice],
                    task_handles=[
                        {"dag_id": "weekly_dag", "task_id": "task1"},
                        {"dag_id": "daily_dag", "task_id": "task1"},
                    ],
                ),
            ],
        ),
    )

    Definitions.validate_loadable(defs)


def test_mixed_multiple_tasks_using_task_mappings() -> None:
    @asset
    def single_targeted_asset() -> None: ...

    @asset
    def double_targeted_asset() -> None: ...

    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            {"weekly_dag": ["task1"], "daily_dag": ["task1"], "other_dag": ["task1"]}
        ),
        defs=Definitions.merge(
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="other_dag", task_mappings={"task1": [single_targeted_asset]}
                )
            ),
            Definitions(
                assets=assets_with_multiple_task_mappings(
                    assets=[double_targeted_asset],
                    task_handles=[
                        {"dag_id": "weekly_dag", "task_id": "task1"},
                        {"dag_id": "daily_dag", "task_id": "task1"},
                    ],
                )
            ),
        ),
    )

    Definitions.validate_loadable(defs)

    mapping_info = build_airlift_metadata_mapping_info(defs.assets)  # type: ignore
    assert mapping_info.all_mapped_asset_keys_by_dag_id["other_dag"] == {
        AssetKey("single_targeted_asset"),
    }
    assert mapping_info.all_mapped_asset_keys_by_dag_id["weekly_dag"] == {
        AssetKey("double_targeted_asset"),
    }
    assert mapping_info.all_mapped_asset_keys_by_dag_id["daily_dag"] == {
        AssetKey("double_targeted_asset"),
    }

    assert mapping_info.task_handle_map[AssetKey("single_targeted_asset")] == {
        TaskHandle(dag_id="other_dag", task_id="task1")
    }
    assert mapping_info.task_handle_map[AssetKey("double_targeted_asset")] == {
        TaskHandle(dag_id="weekly_dag", task_id="task1"),
        TaskHandle(dag_id="daily_dag", task_id="task1"),
    }


def test_task_mappings_in_same_dags() -> None:
    @asset
    def other_asset() -> None: ...

    @asset
    def double_targeted_asset() -> None: ...

    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            {
                "weekly_dag": ["task1", "task_for_other_asset"],
                "daily_dag": ["task1"],
            }
        ),
        defs=Definitions.merge(
            Definitions(
                assets=[
                    *assets_with_multiple_task_mappings(
                        assets=[double_targeted_asset],
                        task_handles=[
                            {"dag_id": "weekly_dag", "task_id": "task1"},
                            {"dag_id": "daily_dag", "task_id": "task1"},
                        ],
                    )
                ]
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="weekly_dag",
                    task_mappings={"task_for_other_asset": [other_asset]},
                ),
            ),
        ),
    )

    Definitions.validate_loadable(defs)

    mapping_info = build_airlift_metadata_mapping_info(defs.assets)  # type: ignore
    assert mapping_info.all_mapped_asset_keys_by_dag_id["weekly_dag"] == {
        AssetKey("other_asset"),
        AssetKey("double_targeted_asset"),
    }
    assert mapping_info.all_mapped_asset_keys_by_dag_id["daily_dag"] == {
        AssetKey("double_targeted_asset"),
    }

    assert mapping_info.task_handle_map[AssetKey("other_asset")] == {
        TaskHandle(dag_id="weekly_dag", task_id="task_for_other_asset")
    }
    assert mapping_info.task_handle_map[AssetKey("double_targeted_asset")] == {
        TaskHandle(dag_id="weekly_dag", task_id="task1"),
        TaskHandle(dag_id="daily_dag", task_id="task1"),
    }


def test_task_mappings_with_same_task_id() -> None:
    @asset
    def other_asset() -> None: ...

    @asset
    def double_targeted_asset() -> None: ...

    defs = build_defs_from_airflow_instance(
        airflow_instance=make_instance(
            {
                "weekly_dag": ["task1"],
                "daily_dag": ["task1"],
            }
        ),
        defs=Definitions.merge(
            Definitions(
                assets=assets_with_multiple_task_mappings(
                    assets=[double_targeted_asset],
                    task_handles=[
                        {"dag_id": "weekly_dag", "task_id": "task1"},
                        {"dag_id": "daily_dag", "task_id": "task1"},
                    ],
                )
            ),
            Definitions(
                assets=assets_with_task_mappings(
                    dag_id="weekly_dag",
                    task_mappings={"task1": [other_asset]},
                ),
            ),
        ),
    )

    Definitions.validate_loadable(defs)

    mapping_info = build_airlift_metadata_mapping_info(defs.assets)  # type: ignore
    assert mapping_info.all_mapped_asset_keys_by_dag_id["weekly_dag"] == {
        AssetKey("other_asset"),
        AssetKey("double_targeted_asset"),
    }
    assert mapping_info.all_mapped_asset_keys_by_dag_id["daily_dag"] == {
        AssetKey("double_targeted_asset"),
    }

    assert mapping_info.task_handle_map[AssetKey("other_asset")] == {
        TaskHandle(dag_id="weekly_dag", task_id="task1")
    }
    assert mapping_info.task_handle_map[AssetKey("double_targeted_asset")] == {
        TaskHandle(dag_id="weekly_dag", task_id="task1"),
        TaskHandle(dag_id="daily_dag", task_id="task1"),
    }


def test_double_instance() -> None:
    airflow_instance_one = make_instance(
        dag_and_task_structure={"dag1": ["task1"]},
        instance_name="instance_one",
    )

    airflow_instance_two = make_instance(
        dag_and_task_structure={"dag1": ["task1"]},
        instance_name="instance_two",
    )

    defs_one = build_defs_from_airflow_instance(airflow_instance=airflow_instance_one)
    defs_two = build_defs_from_airflow_instance(airflow_instance=airflow_instance_two)

    defs = Definitions.merge(defs_one, defs_two)

    all_specs = {spec.key: spec for spec in defs.resolve_all_asset_specs()}

    assert set(all_specs.keys()) == {
        make_default_dag_asset_key("instance_one", "dag1"),
        make_default_dag_asset_key("instance_two", "dag1"),
    }


def test_enrich() -> None:
    spec = AssetSpec(key="a", metadata=metadata_for_task_mapping(task_id="task", dag_id="dag"))
    airflow_assets = enrich_airflow_mapped_assets(
        airflow_instance=make_instance({"dag": ["task"]}),
        mapped_assets=[spec],
        source_code_retrieval_enabled=None,
    )
    assert len(airflow_assets) == 1
    spec = next(iter(airflow_assets))
    assert isinstance(spec, AssetSpec)
    assert spec.key == AssetKey("a")
    # Asset metadata properties have been glommed onto the asset
    assert spec.metadata["Dag ID"] == "dag"


def test_load_dags() -> None:
    dag_assets = load_airflow_dag_asset_specs(
        airflow_instance=make_instance({"dag": ["task"]}),
    )
    assert len(dag_assets) == 1
    dag_asset = next(iter(dag_assets))
    assert dag_asset.key == make_default_dag_asset_key("test_instance", "dag")


def test_load_dags_upstream() -> None:
    upstream_task_spec = AssetSpec(
        key="a", metadata=metadata_for_task_mapping(task_id="task", dag_id="dag")
    )
    dag_assets = load_airflow_dag_asset_specs(
        airflow_instance=make_instance({"dag": ["task"]}),
        mapped_assets=[upstream_task_spec],
    )
    assert len(dag_assets) == 1
    dag_asset_spec = next(iter(dag_assets))
    assert dag_asset_spec.key == make_default_dag_asset_key("test_instance", "dag")
    assert len(list(dag_asset_spec.deps)) == 1
    assert next(iter(dag_asset_spec.deps)).asset_key == AssetKey("a")


def test_filtering() -> None:
    """Test using the retrieval filter to include/exclude dags."""
    instance = make_instance({"include_dag": ["task"], "exclude_dag": ["task"]})
    dag_assets = load_airflow_dag_asset_specs(
        airflow_instance=instance,
        retrieval_filter=AirflowFilter(dag_id_ilike="include"),
    )
    assert len(dag_assets) == 1
    # test tag based retrieval
    instance = make_instance(
        {"dag1": ["task"], "dag2": ["task"]},
        dag_props={"dag1": {"tags": ["first"]}, "dag2": {"tags": ["first", "second"]}},
    )
    dag_assets = load_airflow_dag_asset_specs(
        airflow_instance=instance,
        retrieval_filter=AirflowFilter(airflow_tags=["first"]),
    )
    assert len(dag_assets) == 2
    dag_assets = load_airflow_dag_asset_specs(
        airflow_instance=instance,
        retrieval_filter=AirflowFilter(airflow_tags=["first", "second"]),
    )
    assert len(dag_assets) == 1


def test_load_datasets() -> None:
    """Test automatic loading of datasets."""
    task_structure = {
        "producer1": ["producing_task"],
        "producer2": ["producing_task"],
        "consumer1": ["task"],
        "consumer2": ["task"],
    }
    # Dataset is produced and consumed by multiple tasks
    dataset_info = [
        {
            "uri": "s3://dataset-bucket/example1.csv",
            "producing_tasks": [
                {"dag_id": "producer1", "task_id": "producing_task"},
                {"dag_id": "producer2", "task_id": "producing_task"},
            ],
            "consuming_dags": ["consumer1", "consumer2"],
        },
        {
            "uri": "s3://dataset-bucket/example2.csv",
            "producing_tasks": [
                {"dag_id": "consumer1", "task_id": "task"},
            ],
            "consuming_dags": [],
        },
    ]
    af_instance = make_instance(
        dag_and_task_structure=task_structure,
        dataset_construction_info=dataset_info,
    )
    # Add an additional spec to the same task as the dataset
    spec = AssetSpec(
        key="a", metadata=metadata_for_task_mapping(task_id="producing_task", dag_id="producer1")
    )

    defs = build_defs_from_airflow_instance(
        airflow_instance=af_instance,
        defs=Definitions(assets=[spec]),
    )
    Definitions.validate_loadable(defs)

    definitions_data = AirflowDefinitionsData(
        airflow_instance=af_instance,
        resolved_repository=defs.get_repository_def(),
    )
    assert definitions_data.mapped_asset_keys_by_task_handle == {
        TaskHandle(dag_id="producer1", task_id="producing_task"): {
            AssetKey("example1"),
            AssetKey("a"),
        },
        TaskHandle(dag_id="producer2", task_id="producing_task"): {AssetKey("example1")},
        TaskHandle(dag_id="consumer1", task_id="task"): {AssetKey("example2")},
    }
    example1_spec = asset_spec("example1", defs)
    assert example1_spec
    assert example1_spec.deps == []
    example2_spec = asset_spec("example2", defs)
    assert example2_spec
    assert example2_spec.deps == [AssetDep("example1")]

    # Filter down to just producer1. Only example1 should be included
    defs = build_defs_from_airflow_instance(
        airflow_instance=af_instance,
        retrieval_filter=AirflowFilter(dag_id_ilike="producer1"),
        defs=Definitions(assets=[spec]),
    )
    Definitions.validate_loadable(defs)
    assert asset_spec("example1", defs)
    assert not asset_spec("example2", defs)


def test_load_job_defs() -> None:
    """Test job-based loader."""
    task_structure = {
        "producer1": ["producing_task"],
        "producer2": ["producing_task"],
        "consumer1": ["task"],
        "consumer2": ["task"],
    }
    # Dataset is produced and consumed by multiple tasks
    dataset_info = [
        {
            "uri": "s3://dataset-bucket/example1.csv",
            "producing_tasks": [
                {"dag_id": "producer1", "task_id": "producing_task"},
                {"dag_id": "producer2", "task_id": "producing_task"},
            ],
            "consuming_dags": ["consumer1", "consumer2"],
        },
        {
            "uri": "s3://dataset-bucket/example2.csv",
            "producing_tasks": [
                {"dag_id": "consumer1", "task_id": "task"},
            ],
            "consuming_dags": [],
        },
    ]
    af_instance = make_instance(
        dag_and_task_structure=task_structure,
        dataset_construction_info=dataset_info,
    )
    # Add an additional spec to the same task as the dataset
    spec = AssetSpec(
        key="a", metadata=metadata_for_task_mapping(task_id="producing_task", dag_id="producer1")
    )

    # Add an additional materializable asset to the same task
    @asset(metadata=metadata_for_task_mapping(task_id="producing_task", dag_id="producer1"))
    def b():
        pass

    defs = build_job_based_airflow_defs(
        airflow_instance=af_instance,
        mapped_defs=Definitions(assets=[spec, b]),
    )
    Definitions.validate_loadable(defs)
    assert isinstance(get_job_from_defs("producer1", defs), JobDefinition)
    assert isinstance(get_job_from_defs("producer2", defs), JobDefinition)
    assert isinstance(get_job_from_defs("consumer1", defs), JobDefinition)
    assert isinstance(get_job_from_defs("consumer2", defs), JobDefinition)

    airflow_defs_data = AirflowDefinitionsData(
        airflow_instance=af_instance,
        resolved_repository=defs.get_repository_def(),
    )

    repo = defs.get_repository_def()

    assert airflow_defs_data.airflow_mapped_jobs_by_dag_handle == {
        DagHandle(dag_id="producer1"): repo.get_job("producer1"),
        DagHandle(dag_id="producer2"): repo.get_job("producer2"),
        DagHandle(dag_id="consumer1"): repo.get_job("consumer1"),
        DagHandle(dag_id="consumer2"): repo.get_job("consumer2"),
    }
    assert airflow_defs_data.assets_per_job == {
        "producer1": {AssetKey("example1"), AssetKey("a"), AssetKey("b")},
        "producer2": {AssetKey("example1")},
        "consumer1": {AssetKey("example2")},
        "consumer2": set(),
    }



import time
from pathlib import Path

import pytest
import requests
from dagster import AssetKey, DagsterInstance, DagsterRunStatus
from dagster._core.test_utils import environ
from dagster._time import get_current_timestamp
from dagster_airlift.constants import DAG_RUN_ID_TAG_KEY


def _test_project_dir() -> Path:
    return Path(__file__).parent / "materialize_assets_operator_test_project"


@pytest.fixture(name="dags_dir")
def dags_dir() -> Path:
    return _test_project_dir() / "dags"


@pytest.fixture(name="dagster_defs_path")
def dagster_defs_path_fixture() -> str:
    return str(_test_project_dir() / "dagster_defs.py")


def test_dagster_operator(airflow_instance: None, dagster_dev: None, dagster_home: str) -> None:
    """Tests that dagster operator can correctly map airflow tasks to dagster tasks, and kick off executions."""
    response = requests.post(
        "http://localhost:8080/api/v1/dags/the_dag/dagRuns", auth=("admin", "admin"), json={}
    )
    assert response.status_code == 200, response.json()
    # Wait until the run enters a terminal state
    terminal_status = None
    start_time = get_current_timestamp()
    dag_run = None
    while get_current_timestamp() - start_time < 30:
        response = requests.get(
            "http://localhost:8080/api/v1/dags/the_dag/dagRuns", auth=("admin", "admin")
        )
        assert response.status_code == 200, response.json()
        dag_runs = response.json()["dag_runs"]
        if dag_runs[0]["state"] in ["success", "failed"]:
            terminal_status = dag_runs[0]["state"]
            dag_run = dag_runs[0]
            break
        time.sleep(1)
    assert terminal_status == "success", (
        "Never reached terminal status"
        if terminal_status is None
        else f"terminal status was {terminal_status}"
    )
    with environ({"DAGSTER_HOME": dagster_home}):
        instance = DagsterInstance.get()
        runs = instance.get_runs()
        # The graphql endpoint kicks off a run for each of the assets provided in the asset_key_paths. There are two assets,
        # but they exist within the same job, so there should only be 1 run.
        assert len(runs) == 1
        the_run = [  # noqa
            run
            for run in runs
            if set(list(run.asset_selection))  # type: ignore
            == {
                AssetKey(["some_asset"]),
                AssetKey(["other_asset"]),
                AssetKey(["nested", "asset"]),
                AssetKey(["string", "interpretation"]),
                AssetKey(["backslash/interpretation"]),
            }
        ][0]
        assert the_run.status == DagsterRunStatus.SUCCESS

        assert isinstance(dag_run, dict)
        assert "dag_run_id" in dag_run
        assert the_run.tags[DAG_RUN_ID_TAG_KEY] == dag_run["dag_run_id"]



from datetime import datetime

from airflow import DAG
from airflow.operators.python import PythonOperator


def print_hello() -> None:
    print("Hello")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 1,
}

with DAG(
    "simple_unproxied_dag",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
    tags=["example"],
) as the_dag:
    PythonOperator(task_id="print_task", python_callable=print_hello) >> PythonOperator(
        task_id="downstream_print_task", python_callable=print_hello
    )  # type: ignore



from pathlib import Path

CONSTANTS_FILE = Path(__file__).parent / "constants.txt"


def read_constants() -> dict[str, int]:
    constants = {}
    with open(CONSTANTS_FILE) as f:
        for line in f.readlines():
            key, value = line.strip().split(None, 1)
            constants[key] = int(value)
    return constants


def get_num_dags() -> int:
    constants = read_constants()
    if "NUM_DAGS" not in constants:
        raise ValueError("NUM_DAGS not found in constants file")
    return constants["NUM_DAGS"]


def get_num_tasks() -> int:
    constants = read_constants()
    if "NUM_TASKS" not in constants:
        raise ValueError("NUM_TASKS not found in constants file")
    return constants["NUM_TASKS"]


def get_num_assets() -> int:
    constants = read_constants()
    if "NUM_ASSETS_PER_TASK" not in constants:
        raise ValueError("NUM_ASSETS_PER_TASK not found in constants file")
    return constants["NUM_ASSETS_PER_TASK"]


def get_perf_output_file() -> Path:
    pth = Path(__file__).parent / f"{get_num_dags()}_dags_{get_num_tasks()}_tasks_perf_output.txt"
    pth.parent.mkdir(parents=True, exist_ok=True)
    return pth



import os
import subprocess
import time
from collections.abc import Generator, Mapping, Sequence
from datetime import timedelta
from pathlib import Path
from typing import NamedTuple

import pytest
from dagster import AssetKey, DagsterInstance
from dagster._core.events.log import EventLogEntry
from dagster._core.test_utils import environ
from dagster._time import get_current_datetime
from dagster_airlift.constants import DAG_RUN_ID_TAG_KEY
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.test.shared_fixtures import stand_up_airflow


def makefile_dir() -> Path:
    return Path(__file__).parent.parent.parent


@pytest.fixture(name="local_env")
def local_env_fixture() -> Generator[None, None, None]:
    subprocess.run(["make", "setup_local_env"], cwd=makefile_dir(), check=True)
    with environ(
        {
            "AIRFLOW_HOME": str(makefile_dir() / ".airflow_home"),
            "DAGSTER_HOME": str(makefile_dir() / ".dagster_home"),
        }
    ):
        yield

    subprocess.run(
        ["make", "wipe"],
        cwd=makefile_dir(),
        check=False,  # fail open on cleanup issues
    )


@pytest.fixture(name="dags_dir")
def dags_dir_fixture() -> Path:
    return makefile_dir() / "kitchen_sink" / "airflow_dags"


@pytest.fixture(name="airflow_home")
def airflow_home_fixture(local_env: None) -> Path:
    return Path(os.environ["AIRFLOW_HOME"])


@pytest.fixture(name="dagster_home")
def dagster_home_fixture(local_env: None) -> str:
    return os.environ["DAGSTER_HOME"]


@pytest.fixture(name="expected_num_dags")
def expected_num_dags_fixture() -> int:
    return 1


@pytest.fixture(name="airflow_instance")
def airflow_instance_fixture(
    local_env: None, expected_num_dags: int
) -> Generator[subprocess.Popen, None, None]:
    with stand_up_airflow(
        airflow_cmd=["make", "run_airflow"],
        env=os.environ,
        cwd=makefile_dir(),
        expected_num_dags=expected_num_dags,
    ) as process:
        yield process


def poll_for_airflow_run_existence_and_completion(
    af_instance: AirflowInstance, dag_id: str, af_run_id: str, duration: int
) -> None:
    start_time = time.time()
    while time.time() - start_time < duration:
        try:
            af_instance.wait_for_run_completion(
                dag_id=dag_id, run_id=af_run_id, timeout=int(time.time() - start_time)
            )
            return
        # Run may not exist yet
        except Exception:
            time.sleep(0.1)
            continue


class ExpectedMat(NamedTuple):
    asset_key: AssetKey
    runs_in_dagster: bool


def poll_for_expected_mats(
    af_instance: AirflowInstance,
    expected_mats_per_dag: Mapping[str, Sequence[ExpectedMat | AssetKey]],
) -> None:
    resolved_expected_mats_per_dag: Mapping[str, list[ExpectedMat]] = {
        dag_id: [
            expected_mat
            if isinstance(expected_mat, ExpectedMat)
            else ExpectedMat(expected_mat, True)
            for expected_mat in expected_mats
        ]
        for dag_id, expected_mats in expected_mats_per_dag.items()
    }
    for dag_id, expected_mats in resolved_expected_mats_per_dag.items():
        airflow_run_id = af_instance.trigger_dag(dag_id=dag_id)
        af_instance.wait_for_run_completion(dag_id=dag_id, run_id=airflow_run_id, timeout=60)
        dagster_instance = DagsterInstance.get()

        dag_asset_key = AssetKey([af_instance.name, "dag", dag_id])
        assert poll_for_materialization(dagster_instance, dag_asset_key)

        for expected_mat in expected_mats:
            mat_event_log_entry = poll_for_materialization(dagster_instance, expected_mat.asset_key)
            assert mat_event_log_entry.asset_materialization
            assert mat_event_log_entry.asset_materialization.asset_key == expected_mat.asset_key

            assert mat_event_log_entry.asset_materialization
            dagster_run_id = mat_event_log_entry.run_id

            all_materializations = dagster_instance.fetch_materializations(
                records_filter=expected_mat.asset_key, limit=10
            )

            assert all_materializations

            if expected_mat.runs_in_dagster:
                assert dagster_run_id
                dagster_run = dagster_instance.get_run_by_id(dagster_run_id)
                assert dagster_run
                run_ids = dagster_instance.get_run_ids()
                assert dagster_run, (
                    f"Could not find dagster run {dagster_run_id} All run_ids {run_ids}"
                )
                assert DAG_RUN_ID_TAG_KEY in dagster_run.tags, (
                    f"Could not find dagster run tag: dagster_run.tags {dagster_run.tags}"
                )
                assert dagster_run.tags[DAG_RUN_ID_TAG_KEY] == airflow_run_id, (
                    "dagster run tag does not match dag run id"
                )


def poll_for_materialization(
    dagster_instance: DagsterInstance,
    asset_key: AssetKey,
) -> EventLogEntry:
    start_time = get_current_datetime()
    while get_current_datetime() - start_time < timedelta(seconds=30):
        asset_materialization = dagster_instance.get_latest_materialization_event(
            asset_key=asset_key
        )

        time.sleep(0.1)
        if asset_materialization:
            return asset_materialization

    raise Exception(f"Timeout waiting for materialization event on {asset_key}")



from collections import defaultdict
from collections.abc import Generator, Mapping, Sequence
from datetime import datetime, timedelta
from typing import Any

import pytest
from dagster import (
    AssetKey,
    AssetObservation,
    AssetSpec,
    DagsterInstance,
    Definitions,
    SensorEvaluationContext,
    SensorResult,
    build_sensor_context,
    multi_asset,
)
from dagster._core.definitions.asset_checks.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.definitions_load_context import (
    DefinitionsLoadContext,
    DefinitionsLoadType,
)
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.test_utils import instance_for_test
from dagster._time import get_current_datetime
from dagster_airlift.core import (
    build_defs_from_airflow_instance as build_defs_from_airflow_instance,
)
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.runtime_representations import DagRun, TaskInstance
from dagster_airlift.core.sensor.event_translation import (
    DagsterEventTransformerFn,
    default_event_transformer,
)
from dagster_airlift.core.utils import metadata_for_dag_mapping, metadata_for_task_mapping
from dagster_airlift.test import make_dag_run, make_instance


def strip_to_first_of_month(dt: datetime) -> datetime:
    return dt.replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def fully_loaded_repo_from_airflow_asset_graph(
    assets_per_task: dict[str, dict[str, list[tuple[str, list[str]]]]],
    dataset_construction_info: list[dict[str, Any]] | None = None,
    additional_defs: Definitions = Definitions(),
    create_runs: bool = True,
    dag_level_asset_overrides: dict[str, list[str]] | None = None,
    event_transformer_fn: DagsterEventTransformerFn = default_event_transformer,
) -> RepositoryDefinition:
    defs = load_definitions_airflow_asset_graph(
        assets_per_task,
        dataset_construction_info=dataset_construction_info,
        additional_defs=additional_defs,
        create_runs=create_runs,
        dag_level_asset_overrides=dag_level_asset_overrides,
        event_transformer_fn=event_transformer_fn,
    )
    repo_def = defs.get_repository_def()
    repo_def.load_all_definitions()
    return repo_def


def create_defs_and_instance(
    assets_per_task: dict[str, dict[str, list[tuple[str, list[str]]]]],
    dataset_construction_info: list[dict[str, Any]] | None = None,
    additional_defs: Definitions = Definitions(),
    create_runs: bool = True,
    create_assets_defs: bool = True,
    dag_level_asset_overrides: dict[str, list[str]] | None = None,
    seeded_runs: list[DagRun] | None = None,
    seeded_task_instances: list[TaskInstance] | None = None,
    seeded_logs: Mapping[str, Mapping[str, str]] | None = None,
) -> tuple[Definitions, AirflowInstance]:
    assets = []
    dag_and_task_structure = defaultdict(list)
    for dag_id, task_structure in assets_per_task.items():
        for task_id, asset_structure in task_structure.items():
            dag_and_task_structure[dag_id].append(task_id)
            for asset_key, deps in asset_structure:
                spec = AssetSpec(
                    asset_key,
                    deps=deps,
                    metadata=metadata_for_task_mapping(dag_id=dag_id, task_id=task_id),
                )
                if create_assets_defs:

                    @multi_asset(specs=[spec], name=f"{spec.key.to_python_identifier()}_asset")
                    def _asset():
                        return None

                    assets.append(_asset)
                else:
                    assets.append(spec)
    if dag_level_asset_overrides:
        for dag_id, asset_keys in dag_level_asset_overrides.items():
            dag_and_task_structure[dag_id] = ["dummy_task"]
            for asset in asset_keys:
                spec = AssetSpec(
                    AssetKey.from_user_string(asset),
                    metadata=metadata_for_dag_mapping(dag_id=dag_id),
                )
                if create_assets_defs:

                    @multi_asset(specs=[spec], name=f"{spec.key.to_python_identifier()}_asset")
                    def _asset():
                        return None

                    assets.append(_asset)
                else:
                    assets.append(spec)
    if not seeded_runs:
        runs = (
            [
                make_dag_run(
                    dag_id=dag_id,
                    run_id=f"run-{dag_id}",
                    start_date=get_current_datetime() - timedelta(minutes=10),
                    end_date=get_current_datetime(),
                )
                for dag_id in dag_and_task_structure.keys()
            ]
            if create_runs
            else []
        )
    else:
        runs = seeded_runs
    instance = make_instance(
        dag_and_task_structure=dag_and_task_structure,
        dag_runs=runs,
        dataset_construction_info=dataset_construction_info or [],
        task_instances=seeded_task_instances,
        logs=seeded_logs,
    )
    defs = Definitions.merge(
        additional_defs,
        Definitions(assets=assets),
    )
    return defs, instance


def load_definitions_airflow_asset_graph(
    assets_per_task: dict[str, dict[str, list[tuple[str, list[str]]]]],
    dataset_construction_info: list[dict[str, Any]] | None = None,
    additional_defs: Definitions = Definitions(),
    create_runs: bool = True,
    create_assets_defs: bool = True,
    dag_level_asset_overrides: dict[str, list[str]] | None = None,
    event_transformer_fn: DagsterEventTransformerFn = default_event_transformer,
    seeded_runs: list[DagRun] | None = None,
    seeded_task_instances: list[TaskInstance] | None = None,
) -> Definitions:
    defs, instance = create_defs_and_instance(
        assets_per_task=assets_per_task,
        dataset_construction_info=dataset_construction_info,
        additional_defs=additional_defs,
        create_runs=create_runs,
        create_assets_defs=create_assets_defs,
        dag_level_asset_overrides=dag_level_asset_overrides,
        seeded_runs=seeded_runs,
        seeded_task_instances=seeded_task_instances,
    )
    return build_defs_from_airflow_instance(
        airflow_instance=instance, defs=defs, event_transformer_fn=event_transformer_fn
    )


def build_and_invoke_sensor(
    *,
    assets_per_task: dict[str, dict[str, list[tuple[str, list[str]]]]],
    instance: DagsterInstance,
    additional_defs: Definitions = Definitions(),
    dag_level_asset_overrides: dict[str, list[str]] | None = None,
    event_transformer_fn: DagsterEventTransformerFn = default_event_transformer,
) -> tuple[SensorResult, SensorEvaluationContext]:
    repo_def = fully_loaded_repo_from_airflow_asset_graph(
        assets_per_task,
        additional_defs=additional_defs,
        dag_level_asset_overrides=dag_level_asset_overrides,
        event_transformer_fn=event_transformer_fn,
    )
    sensor = next(iter(repo_def.sensor_defs))
    sensor_context = build_sensor_context(repository_def=repo_def, instance=instance)
    result = sensor(sensor_context)
    assert isinstance(result, SensorResult)
    return result, sensor_context


def assert_expected_key_order(
    mats: Sequence[AssetMaterialization | AssetObservation | AssetCheckEvaluation],
    expected_key_order: Sequence[str],
) -> None:
    assert all(isinstance(mat, AssetMaterialization) for mat in mats)
    assert [mat.asset_key.to_user_string() for mat in mats] == expected_key_order


def assert_dependency_structure_in_assets(
    repo_def: RepositoryDefinition, expected_deps: dict[str, list[str]]
) -> None:
    for key, deps_list in expected_deps.items():
        qual_key = AssetKey.from_user_string(key)
        assert qual_key in repo_def.assets_defs_by_key
        assets_def = repo_def.assets_defs_by_key[qual_key]
        spec = next(spec for spec in assets_def.specs if spec.key == qual_key)
        assert {dep.asset_key for dep in spec.deps} == {
            AssetKey.from_user_string(dep) for dep in deps_list
        }


@pytest.fixture(name="instance")
def instance_fixture() -> Generator[DagsterInstance, None, None]:
    with instance_for_test() as instance:
        yield instance


def _set_init_load_context() -> None:
    """Sets the load context to initialization."""
    DefinitionsLoadContext.set(DefinitionsLoadContext(load_type=DefinitionsLoadType.INITIALIZATION))


@pytest.fixture(name="init_load_context")
def load_context_fixture() -> Generator[None, None, None]:
    """Set initialization load context before and after the test."""
    _set_init_load_context()
    yield
    _set_init_load_context()



from collections.abc import Sequence
from typing import TYPE_CHECKING, cast

import pytest
import requests
from dagster import Definitions
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.run_request import SensorResult
from dagster._core.definitions.sensor_definition import build_sensor_context
from dagster._core.test_utils import instance_for_test
from dagster_airlift.constants import PEERED_DAG_MAPPING_METADATA_KEY, SOURCE_CODE_METADATA_KEY
from dagster_airlift.core import build_defs_from_airflow_instance
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.basic_auth import AirflowBasicAuthBackend
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.serialization.serialized_data import Dataset
from dagster_airlift.core.top_level_dag_def_api import assets_with_dag_mappings
from dagster_airlift.test.test_utils import asset_spec
from kitchen_sink.airflow_instance import (
    AIRFLOW_BASE_URL,
    AIRFLOW_INSTANCE_NAME,
    EXPECTED_NUM_DAGS,
    PASSWORD,
    USERNAME,
    local_airflow_instance,
)
from pytest_mock import MockFixture

if TYPE_CHECKING:
    from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition


@pytest.fixture
def expected_num_dags() -> int:
    return EXPECTED_NUM_DAGS


def test_configure_dag_list_limit(airflow_instance: None, mocker: MockFixture) -> None:
    """Test that batch instance logic correctly retrieves all dags when over batch limit."""
    spy = mocker.spy(requests.Session, "get")
    af_instance = AirflowInstance(
        auth_backend=AirflowBasicAuthBackend(
            webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
        ),
        name=AIRFLOW_INSTANCE_NAME,
        # Set low list limit, force batched retrieval.
        dag_list_limit=1,
    )
    assert len(af_instance.list_dags(AirflowFilter())) == EXPECTED_NUM_DAGS
    # EXPECTED_NUM_DAGS with actual results, 1 with no results
    assert spy.call_count == EXPECTED_NUM_DAGS + 1


def test_airflow_filter(airflow_instance: None) -> None:
    """Test that airflow filter correctly filters dags."""
    af_instance = AirflowInstance(
        auth_backend=AirflowBasicAuthBackend(
            webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
        ),
        name=AIRFLOW_INSTANCE_NAME,
    )
    dags = af_instance.list_dags(AirflowFilter(dag_id_ilike="simple"))
    assert len(dags) == 1
    assert dags[0].dag_id == "simple_unproxied_dag"

    dags = af_instance.list_dags(AirflowFilter(airflow_tags=["example"]))
    assert len(dags) == 1
    assert dags[0].dag_id == "simple_unproxied_dag"


def change_dag_limit_source_code(limit: int) -> None:
    import dagster_airlift.core.serialization.compute

    dagster_airlift.core.serialization.compute.DEFAULT_MAX_NUM_DAGS_SOURCE_CODE_RETRIEVAL = limit


def test_disable_source_code_retrieval_at_scale(airflow_instance: None) -> None:
    """Test that source code retrieval is disabled at scale."""
    change_dag_limit_source_code(1)
    af_instance = local_airflow_instance()
    defs = build_defs_from_airflow_instance(airflow_instance=af_instance)
    assert defs.assets
    for assets_def in defs.assets:
        metadata = next(iter(cast("AssetsDefinition", assets_def).specs)).metadata
        assert SOURCE_CODE_METADATA_KEY not in metadata

    change_dag_limit_source_code(200)
    # Also force re-retrieval of state by giving the airflow instance a different name.
    af_instance = local_airflow_instance(name="different_name")
    defs = build_defs_from_airflow_instance(airflow_instance=af_instance)
    assert defs.assets
    for assets_def in defs.assets:
        metadata = next(iter(cast("AssetsDefinition", assets_def).specs)).metadata
        if PEERED_DAG_MAPPING_METADATA_KEY not in metadata:
            continue
        assert SOURCE_CODE_METADATA_KEY in metadata

    # If source code retrieval is explicitly enabled, we don't use the limit.
    change_dag_limit_source_code(1)
    af_instance = local_airflow_instance(name="different_name")
    defs = build_defs_from_airflow_instance(
        airflow_instance=af_instance, source_code_retrieval_enabled=True
    )
    assert defs.assets
    for assets_def in defs.assets:
        metadata = next(iter(cast("AssetsDefinition", assets_def).specs)).metadata
        if PEERED_DAG_MAPPING_METADATA_KEY not in metadata:
            continue
        assert SOURCE_CODE_METADATA_KEY in metadata


def test_sensor_iteration_multiple_batches(airflow_instance: None, mocker: MockFixture) -> None:
    """Test that sensor iteration correctly retrieves all runs when over batch limit."""
    spy = mocker.spy(AirflowInstance, "get_dag_runs_batch")
    af_instance = AirflowInstance(
        auth_backend=AirflowBasicAuthBackend(
            webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
        ),
        name=AIRFLOW_INSTANCE_NAME,
        # Set low list limit, force batched retrieval.
        batch_dag_runs_limit=1,
    )
    defs = build_defs_from_airflow_instance(airflow_instance=af_instance)
    # create a bunch of runs
    for i in range(2):
        run_id = af_instance.trigger_dag("simple_unproxied_dag")
        af_instance.wait_for_run_completion(dag_id="simple_unproxied_dag", run_id=run_id)

    assert defs.sensors
    sensor_def = next(iter(defs.sensors))
    with instance_for_test() as instance:
        ctx = build_sensor_context(instance=instance, definitions=defs)
        result = sensor_def(ctx)
        assert isinstance(result, SensorResult)
        assert len(result.asset_events) == 2
        assert spy.call_count == 2


def test_sensor_explicitly_mapped_assets(airflow_instance: None, mocker: MockFixture) -> None:
    """Test that sensor iteration correctly retrieves all runs even when assets are explicitly mapped."""
    af_instance = AirflowInstance(
        auth_backend=AirflowBasicAuthBackend(
            webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
        ),
        name=AIRFLOW_INSTANCE_NAME,
    )
    # explicitly map an asset to a dag. This forces the set of dag_ids passed to the rest API to be nonzero.
    defs = build_defs_from_airflow_instance(
        airflow_instance=af_instance,
        defs=Definitions(assets=assets_with_dag_mappings({"print_dag": [AssetSpec("my_asset")]})),
    )
    # create a bunch of runs for an unrelated "peer only" dag
    for i in range(2):
        run_id = af_instance.trigger_dag("simple_unproxied_dag")
        af_instance.wait_for_run_completion(dag_id="simple_unproxied_dag", run_id=run_id)

    assert defs.sensors
    sensor_def = next(iter(defs.sensors))
    with instance_for_test() as instance:
        ctx = build_sensor_context(instance=instance, definitions=defs)
        result = sensor_def(ctx)
        assert isinstance(result, SensorResult)
        assert len(result.asset_events) == 2


def dataset_with_uri(datasets: Sequence[Dataset], uri: str) -> Dataset:
    return next(dataset for dataset in datasets if dataset.uri == uri)


def test_datasets(airflow_instance: None) -> None:
    """Test that we can correctly retrieve datasets from Airflow."""
    af_instance = local_airflow_instance()
    datasets = af_instance.get_all_datasets()
    assert len(datasets) == 2
    assert {d.uri for d in datasets} == {
        "s3://dataset-bucket/example1.csv",
        "s3://dataset-bucket/example2.csv",
    }
    example1_dataset = dataset_with_uri(datasets, "s3://dataset-bucket/example1.csv")
    assert {t.task_id for t in example1_dataset.producing_tasks} == {"print_task"}
    assert {d.dag_id for d in example1_dataset.consuming_dags} == {"example1_consumer"}
    assert example1_dataset.is_produced_by_task(task_id="print_task", dag_id="dataset_producer")

    example2_dataset = dataset_with_uri(datasets, "s3://dataset-bucket/example2.csv")
    assert {t.task_id for t in example2_dataset.producing_tasks} == {"print_task"}
    assert {d.dag_id for d in example2_dataset.consuming_dags} == {"example2_consumer"}
    assert example2_dataset.is_produced_by_task(task_id="print_task", dag_id="dataset_producer")

    # Apply a filter to the dataset
    datasets = af_instance.get_all_datasets(
        retrieval_filter=AirflowFilter(dataset_uri_ilike="example1")
    )
    assert len(datasets) == 1

    defs = build_defs_from_airflow_instance(airflow_instance=af_instance)
    assert asset_spec("example1", defs)
    assert asset_spec("example2", defs)

    # Apply a dag filter that does not include the producing dag. The datasets should not be
    # included in the definitions.
    defs = build_defs_from_airflow_instance(
        airflow_instance=af_instance,
        retrieval_filter=AirflowFilter(dag_id_ilike="print_dag"),
    )
    assert not asset_spec("example1", defs)
    assert not asset_spec("example2", defs)


def test_log_retrieval(airflow_instance: None) -> None:
    af_instance = local_airflow_instance()
    run_id = af_instance.trigger_dag("dataset_producer")
    af_instance.wait_for_run_completion(dag_id="dataset_producer", run_id=run_id)
    logs = af_instance.get_task_instance_logs(
        dag_id="dataset_producer", task_id="print_task", run_id=run_id, try_number=1
    )
    assert logs
    assert "DAGSTER_START" in logs
    assert "DAGSTER_END" in logs



from dagster_airlift.core import AirflowBasicAuthBackend, AirflowInstance

from kitchen_sink.constants import AIRFLOW_BASE_URL, AIRFLOW_INSTANCE_NAME, PASSWORD, USERNAME

EXPECTED_NUM_DAGS = 20


def local_airflow_instance(name: str | None = None) -> AirflowInstance:
    return AirflowInstance(
        auth_backend=AirflowBasicAuthBackend(
            webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
        ),
        name=name or AIRFLOW_INSTANCE_NAME,
    )



from datetime import datetime, timedelta

from airflow import DAG
from airflow.operators.python import PythonOperator


def succeeds_on_retry(**context) -> None:
    if context["task_instance"].try_number == 1:
        raise Exception("Failing on first try")
    else:
        print("Succeeding on retry")  # noqa: T201


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 2,
}


with DAG(
    "unmapped__dag_with_retries",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as dag:
    PythonOperator(
        task_id="print_task", python_callable=succeeds_on_retry, retry_delay=timedelta(seconds=1)
    )



import datetime

import pytest
from dagster import _check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.metadata.metadata_value import TimestampMetadataValue
from dagster._core.definitions.run_config import RunConfig
from dagster._core.event_api import EventRecordsFilter
from dagster._core.events import DagsterEventType
from dagster._core.instance import DagsterInstance
from dagster._core.instance_for_test import instance_for_test
from dagster._core.storage.dagster_run import DagsterRunStatus, RunsFilter
from dagster._core.storage.tags import EXTERNAL_JOB_SOURCE_TAG_KEY, EXTERNALLY_MANAGED_ASSETS_TAG
from dagster_airlift.constants import (
    DAG_ID_TAG_KEY,
    DAG_RUN_ID_TAG_KEY,
    OBSERVATION_RUN_TAG_KEY,
    TASK_ID_TAG_KEY,
)
from dagster_airlift.core.monitoring_job.builder import MonitoringConfig, monitoring_job_op_name
from dagster_airlift.core.utils import monitoring_job_name
from dagster_airlift.test.test_utils import asset_spec
from dagster_test.utils.definitions_execute_in_process import definitions_execute_job_in_process
from kitchen_sink.airflow_instance import local_airflow_instance

from kitchen_sink_tests.integration_tests.conftest import (
    makefile_dir,
    poll_for_airflow_run_existence_and_completion,
)


def test_job_based_defs(
    airflow_instance: None,
) -> None:
    """Test that job based defs load properly."""
    from kitchen_sink.dagster_defs.job_based_defs import defs

    assert len(defs.jobs) == 21  # type: ignore
    assert len(defs.assets) == 4  # type: ignore
    for key in ["print_asset", "another_print_asset", "example1", "example2"]:
        assert asset_spec(key, defs)

    # First, execute dataset producer dag
    af_instance = local_airflow_instance()
    af_run_id = af_instance.trigger_dag("dataset_producer")
    poll_for_airflow_run_existence_and_completion(
        af_instance=af_instance, af_run_id=af_run_id, dag_id="dataset_producer", duration=30
    )

    # Then, execute monitoring job
    with instance_for_test() as instance:
        result = definitions_execute_job_in_process(
            defs=defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(
                            datetime.datetime.now() - datetime.timedelta(seconds=30)
                        ).isoformat(),
                        range_end=datetime.datetime.now().isoformat(),
                    )
                }
            ),
        )
        assert result.success
        # There should be a run for the dataset producer dag and a run for the monitoring job
        runs = instance.get_runs()
        assert len(runs) == 2

        producer_run = next(run for run in runs if run.job_name == "dataset_producer")
        assert producer_run.status == DagsterRunStatus.SUCCESS
        assert producer_run.tags[DAG_RUN_ID_TAG_KEY] == af_run_id
        assert producer_run.tags[DAG_ID_TAG_KEY] == "dataset_producer"
        assert producer_run.tags[EXTERNAL_JOB_SOURCE_TAG_KEY] == "airflow"

        # Check that there are asset planned events for the two assets
        planned_records = instance.get_event_records(
            event_records_filter=EventRecordsFilter(
                event_type=DagsterEventType.ASSET_MATERIALIZATION_PLANNED
            ),
        )
        assert len(planned_records) == 2
        assert {planned_records.asset_key for planned_records in planned_records} == {
            AssetKey("example1"),
            AssetKey("example2"),
        }
        assert {planned_records.run_id for planned_records in planned_records} == {
            producer_run.run_id
        }

        # Check that there are asset materialized events for the two assets
        materialized_records = instance.get_event_records(
            event_records_filter=EventRecordsFilter(
                event_type=DagsterEventType.ASSET_MATERIALIZATION
            ),
        )
        assert len(materialized_records) == 2
        assert {
            materialized_records.asset_key for materialized_records in materialized_records
        } == {
            AssetKey("example1"),
            AssetKey("example2"),
        }
        assert {materialized_records.run_id for materialized_records in materialized_records} == {
            producer_run.run_id
        }
        for key in ["example1", "example2"]:
            key_record = next(
                materialized_records
                for materialized_records in materialized_records
                if materialized_records.asset_key == AssetKey(key)
            )
            asset_metadata = check.not_none(key_record.asset_materialization).metadata

            assert asset_metadata["my_timestamp"] == TimestampMetadataValue(value=111.0)
            assert asset_metadata["my_other_timestamp"] == TimestampMetadataValue(value=113.0)
            # It gets overridden by the second print
            assert asset_metadata["foo"].value == "baz"


@pytest.fixture(name="dagster_dev_cmd")
def dagster_dev_cmd_fixture() -> list[str]:
    return ["make", "run_job_based_defs", "-C", str(makefile_dir())]


def test_job_based_defs_with_proxied_assets(
    airflow_instance: None,
    dagster_dev: None,
    dagster_home: str,
) -> None:
    from kitchen_sink.dagster_defs.job_based_defs import defs

    af_instance = local_airflow_instance()
    with DagsterInstance.get() as instance:
        # Execute print_dag. We'd expect corresponding runs to be kicked off of print_asset and another_print_asset.
        print_dag_run_id = af_instance.trigger_dag("deferred_events_dag")
        poll_for_airflow_run_existence_and_completion(
            af_instance=af_instance,
            af_run_id=print_dag_run_id,
            dag_id="deferred_events_dag",
            duration=30,
        )

        # First, there should be "asset-less" runs for print_asset and another_print_asset
        print_task_run = next(
            iter(
                instance.get_run_records(
                    filters=RunsFilter(
                        tags={
                            TASK_ID_TAG_KEY: "print_task",
                            DAG_ID_TAG_KEY: "deferred_events_dag",
                            DAG_RUN_ID_TAG_KEY: print_dag_run_id,
                            EXTERNALLY_MANAGED_ASSETS_TAG: "true",
                        }
                    )
                )
            )
        )
        assert print_task_run.dagster_run.status == DagsterRunStatus.SUCCESS

        # Attempt to retrieve materialization events and planned events scoped to the run. They should be empty.
        conn = instance.event_log_storage.get_records_for_run(
            run_id=print_task_run.dagster_run.run_id,
            of_type={
                DagsterEventType.ASSET_MATERIALIZATION,
                DagsterEventType.ASSET_MATERIALIZATION_PLANNED,
            },
        )
        assert len(conn.records) == 0

        # Same for downstream_print_task
        downstream_print_task_run = next(
            iter(
                instance.get_run_records(
                    filters=RunsFilter(
                        tags={
                            TASK_ID_TAG_KEY: "downstream_print_task",
                            DAG_ID_TAG_KEY: "deferred_events_dag",
                            DAG_RUN_ID_TAG_KEY: print_dag_run_id,
                            EXTERNALLY_MANAGED_ASSETS_TAG: "true",
                        }
                    )
                )
            )
        )
        assert downstream_print_task_run.dagster_run.status == DagsterRunStatus.SUCCESS

        # Attempt to retrieve materialization events and planned events scoped to the run. They should be empty.
        conn = instance.event_log_storage.get_records_for_run(
            run_id=downstream_print_task_run.dagster_run.run_id,
            of_type={
                DagsterEventType.ASSET_MATERIALIZATION,
                DagsterEventType.ASSET_MATERIALIZATION_PLANNED,
            },
        )
        assert len(conn.records) == 0

        # Execute the monitoring job in process.
        result = definitions_execute_job_in_process(
            defs=defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(
                            datetime.datetime.now() - datetime.timedelta(seconds=300)
                        ).isoformat(),
                        range_end=datetime.datetime.now().isoformat(),
                    )
                }
            ),
        )
        assert result.success

        # Check that the print_asset was materialized, and that the metadata from the run was included.
        observed_run = next(
            iter(
                instance.get_run_records(
                    filters=RunsFilter(
                        tags={
                            DAG_ID_TAG_KEY: "deferred_events_dag",
                            DAG_RUN_ID_TAG_KEY: print_dag_run_id,
                            OBSERVATION_RUN_TAG_KEY: "true",
                        }
                    )
                )
            )
        )
        assert observed_run.dagster_run.status == DagsterRunStatus.SUCCESS
        conn = instance.event_log_storage.get_records_for_run(
            run_id=observed_run.dagster_run.run_id,
            of_type={DagsterEventType.ASSET_MATERIALIZATION},
        )
        assert len(conn.records) == 2
        materializations = [
            check.not_none(r.event_log_entry.asset_materialization) for r in conn.records
        ]
        assert {m.asset_key for m in materializations} == {
            AssetKey("print_asset"),
            AssetKey("another_print_asset"),
        }
        print_mat = next(m for m in materializations if m.asset_key == AssetKey("print_asset"))
        assert print_mat.metadata["foo"].value == "bar"
        another_print_mat = next(
            m for m in materializations if m.asset_key == AssetKey("another_print_asset")
        )
        assert another_print_mat.metadata["foo"].value == "baz"



import json
from datetime import datetime, timedelta, timezone

import dagster._check as check
import pytest
from dagster import AssetKey, DagsterInstance
from dagster._core.definitions import materialize
from dagster._core.definitions.decorators.asset_decorator import asset
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.metadata.metadata_value import TextMetadataValue
from dagster._core.definitions.run_config import RunConfig
from dagster._core.definitions.run_request import RunRequest, SkipReason
from dagster._core.definitions.sensor_definition import build_sensor_context
from dagster._core.events import DagsterEvent
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.instance_for_test import instance_for_test
from dagster._core.storage.dagster_run import DagsterRun, DagsterRunStatus, RunsFilter
from dagster._core.storage.tags import (
    EXTERNALLY_MANAGED_ASSETS_TAG,
    REPOSITORY_LABEL_TAG,
    ROOT_RUN_ID_TAG,
)
from dagster._core.test_utils import freeze_time
from dagster._core.utils import make_new_run_id
from dagster_airlift.constants import (
    DAG_ID_TAG_KEY,
    DAG_RUN_ID_TAG_KEY,
    OBSERVATION_RUN_TAG_KEY,
    TASK_ID_TAG_KEY,
)
from dagster_airlift.core.job_builder import job_name
from dagster_airlift.core.load_defs import (
    build_defs_from_airflow_instance,
    build_job_based_airflow_defs,
)
from dagster_airlift.core.monitoring_job.builder import (
    MonitoringConfig,
    build_airflow_monitoring_defs,
    monitoring_job_op_name,
)
from dagster_airlift.core.serialization.defs_construction import make_default_dag_asset_key
from dagster_airlift.core.utils import monitoring_job_name
from dagster_airlift.test import make_dag_run, make_task_instance
from dagster_test.utils.definitions_execute_in_process import definitions_execute_job_in_process

from dagster_airlift_tests.unit_tests.conftest import create_defs_and_instance


def make_dag_key(dag_id: str) -> AssetKey:
    return AssetKey(["test_instance", "dag", dag_id])


def make_dag_key_str(dag_id: str) -> str:
    return make_dag_key(dag_id).to_user_string()


def test_monitoring_job_execution(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that monitoring job correctly represents state in Dagster."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)

    raw_metadata = {
        "foo": "bar",
        "my_timestamp": {"raw_value": 111, "type": "timestamp"},
    }
    with freeze_time(freeze_datetime):
        defs, af_instance = create_defs_and_instance(
            assets_per_task={
                "dag": {"task": [("a", [])]},
                "dag2": {"task": [("b", [])]},
            },
            create_runs=False,
            create_assets_defs=False,
            seeded_runs=[
                # Have a newly started run.
                make_dag_run(
                    dag_id="dag",
                    run_id="run-dag",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=None,
                    state="running",
                ),
                # Have a newly finished run in success state. We also need to create a corresponding run on the instance
                # for this.
                make_dag_run(
                    dag_id="dag",
                    run_id="success-run",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                    state="success",
                ),
                # Have a newly finished run in failed state. We also need to create a corresponding run on the instance
                # for this.
                make_dag_run(
                    dag_id="dag",
                    run_id="failure-run",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                    state="failed",
                ),
                # Newly finished run that started after the last iteration, and therefore has no corresponding run on the instance.
                make_dag_run(
                    dag_id="dag2",
                    run_id="late-run",
                    start_date=freeze_datetime - timedelta(seconds=15),
                    end_date=freeze_datetime - timedelta(seconds=10),
                    state="success",
                ),
            ],
            seeded_task_instances=[
                # Have a newly completed task instance for the newly started run.
                make_task_instance(
                    dag_id="dag",
                    task_id="task",
                    run_id="run-dag",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                ),
                # Have a newly completed task instance for the late run.
                make_task_instance(
                    dag_id="dag2",
                    task_id="task",
                    run_id="late-run",
                    start_date=freeze_datetime - timedelta(seconds=15),
                    end_date=freeze_datetime - timedelta(seconds=10),
                ),
            ],
            seeded_logs={
                "run-dag": {"task": f"DAGSTER_START{json.dumps(raw_metadata)}DAGSTER_END"}
            },
        )
        defs = build_job_based_airflow_defs(
            airflow_instance=af_instance,
            mapped_defs=defs,
        )
        success_dagster_run_id = make_new_run_id()
        instance.run_storage.add_historical_run(
            dagster_run=DagsterRun(
                job_name=job_name("dag"),
                run_id=success_dagster_run_id,
                tags={
                    DAG_RUN_ID_TAG_KEY: "success-run",
                    DAG_ID_TAG_KEY: "dag",
                    OBSERVATION_RUN_TAG_KEY: "true",
                },
                status=DagsterRunStatus.STARTED,
            ),
            run_creation_time=freeze_datetime - timedelta(seconds=30),
        )
        failure_dagster_run_id = make_new_run_id()
        instance.run_storage.add_historical_run(
            dagster_run=DagsterRun(
                job_name=job_name("dag"),
                run_id=failure_dagster_run_id,
                tags={
                    DAG_RUN_ID_TAG_KEY: "failure-run",
                    DAG_ID_TAG_KEY: "dag",
                    OBSERVATION_RUN_TAG_KEY: "true",
                },
            ),
            run_creation_time=freeze_datetime - timedelta(seconds=30),
        )
        result = definitions_execute_job_in_process(
            defs=defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            tags={REPOSITORY_LABEL_TAG: "placeholder"},
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(freeze_datetime - timedelta(seconds=30)).isoformat(),
                        range_end=freeze_datetime.isoformat(),
                    )
                }
            ),
        )
        assert result.success

        # Expect that the success and failure runs are marked as finished.
        success_record = check.not_none(instance.get_run_record_by_id(success_dagster_run_id))
        assert success_record.dagster_run.status == DagsterRunStatus.SUCCESS
        assert success_record.end_time == (freeze_datetime).timestamp()

        failure_record = check.not_none(instance.get_run_record_by_id(failure_dagster_run_id))
        assert failure_record.dagster_run.status == DagsterRunStatus.FAILURE
        assert failure_record.end_time == (freeze_datetime).timestamp()

        # Expect that we created a new run for the newly running run.
        newly_started_run_record = next(
            iter(instance.get_run_records(filters=RunsFilter(tags={DAG_RUN_ID_TAG_KEY: "run-dag"})))
        )
        assert newly_started_run_record.dagster_run.status == DagsterRunStatus.STARTED
        assert (
            newly_started_run_record.start_time
            == (freeze_datetime - timedelta(seconds=30)).timestamp()
        )

        late_run = next(
            iter(instance.get_runs(filters=RunsFilter(tags={DAG_RUN_ID_TAG_KEY: "late-run"})))
        )
        assert late_run.status == DagsterRunStatus.SUCCESS
        run_record = check.not_none(instance.get_run_record_by_id(late_run.run_id))
        assert (
            run_record.create_timestamp.timestamp()
            == (freeze_datetime - timedelta(seconds=15)).timestamp()
        )
        assert run_record.start_time == (freeze_datetime - timedelta(seconds=15)).timestamp()
        assert run_record.end_time == (freeze_datetime - timedelta(seconds=10)).timestamp()

        # There should be planned materialization data for the task.
        planned_info = instance.get_latest_planned_materialization_info(AssetKey("a"))
        assert planned_info
        assert planned_info.run_id == newly_started_run_record.dagster_run.run_id
        # Expect that we emitted asset materialization events for the task.
        mapped_asset_mat = instance.get_latest_materialization_event(AssetKey("a"))
        assert mapped_asset_mat is not None
        assert mapped_asset_mat.run_id == newly_started_run_record.dagster_run.run_id


def get_invalid_json_log_content() -> str:
    raw_metadata = {
        "foo": "bar",
        "my_timestamp": {"raw_value": 111, "type": "timestamp"},
    }
    return f"DAGSTER_START{json.dumps(raw_metadata)}ijkDAGSTER_END"  # add ijk randomly at the endto make it invalid json


def get_invalid_type_log_content() -> str:
    raw_metadata = {
        "foo": "bar",
        "my_timestamp": {"raw_value": 111, "type": "not_a_type"},
    }
    return f"DAGSTER_START{json.dumps(raw_metadata)}DAGSTER_END"


@pytest.mark.parametrize(
    "log_content",
    [
        get_invalid_json_log_content(),
        get_invalid_type_log_content(),
    ],
)
def test_monitoring_job_log_extraction_errors(
    init_load_context: None, instance: DagsterInstance, log_content: str
) -> None:
    """Test that monitoring job correctly handles log extraction errors."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)

    with freeze_time(freeze_datetime):
        defs, af_instance = create_defs_and_instance(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            create_runs=False,
            create_assets_defs=False,
            seeded_runs=[
                # Have a newly started run.
                make_dag_run(
                    dag_id="dag",
                    run_id="run-dag",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=None,
                    state="running",
                ),
            ],
            seeded_task_instances=[
                # Have a newly completed task instance for the newly started run.
                make_task_instance(
                    dag_id="dag",
                    task_id="task",
                    run_id="run-dag",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                ),
            ],
            seeded_logs={"run-dag": {"task": log_content}},
        )
        defs = build_job_based_airflow_defs(
            airflow_instance=af_instance,
            mapped_defs=defs,
        )
        # Despite the invalid log content, we should still be able to execute the job.
        result = definitions_execute_job_in_process(
            defs=defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            tags={REPOSITORY_LABEL_TAG: "placeholder"},
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(freeze_datetime - timedelta(seconds=30)).isoformat(),
                        range_end=freeze_datetime.isoformat(),
                    )
                }
            ),
        )
        assert result.success
        newly_started_run = next(
            iter(instance.get_runs(filters=RunsFilter(tags={DAG_RUN_ID_TAG_KEY: "run-dag"})))
        )

        # Expect that we emitted asset materialization events for the task.
        mapped_asset_mat = instance.get_latest_materialization_event(AssetKey("a"))
        assert mapped_asset_mat is not None
        assert mapped_asset_mat.run_id == newly_started_run.run_id
        # metadata should be empty
        assert check.not_none(mapped_asset_mat.asset_materialization).metadata == {}


def test_monitoring_job_dag_assets(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that monitoring job can correctly handle state when dag assets are used."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)

    with freeze_time(freeze_datetime):
        defs, af_instance = create_defs_and_instance(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            create_runs=False,
            create_assets_defs=False,
            seeded_runs=[
                # Have a newly started run.
                make_dag_run(
                    dag_id="dag",
                    run_id="run-dag",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=None,
                    state="running",
                ),
                # Have a newly finished run in success state. Since there's no jobs associated,
                # we don't need to create a corresponding run on the instance.
                make_dag_run(
                    dag_id="dag",
                    run_id="success-run",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                    state="success",
                ),
                # Have a newly finished run in failed state. Since there's no jobs associated,
                # we don't need to create a corresponding run on the instance.
                make_dag_run(
                    dag_id="dag",
                    run_id="failure-run",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                    state="failed",
                ),
            ],
            seeded_task_instances=[
                # Have a newly completed task instance for the newly started run.
                make_task_instance(
                    dag_id="dag",
                    task_id="task",
                    run_id="run-dag",
                    start_date=freeze_datetime - timedelta(seconds=30),
                    end_date=freeze_datetime,
                ),
            ],
        )
        enriched_defs = build_defs_from_airflow_instance(
            airflow_instance=af_instance,
            defs=defs,
        )
        full_defs = Definitions.merge(
            enriched_defs, build_airflow_monitoring_defs(airflow_instance=af_instance)
        )
        result = definitions_execute_job_in_process(
            defs=full_defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            tags={REPOSITORY_LABEL_TAG: "placeholder"},
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(freeze_datetime - timedelta(seconds=30)).isoformat(),
                        range_end=freeze_datetime.isoformat(),
                    )
                }
            ),
        )
        assert result.success
        # There should be runless materializations for the dag asset corresponding to run-dag,
        dag_mat = instance.get_latest_materialization_event(
            make_default_dag_asset_key(instance_name=af_instance.name, dag_id="dag")
        )
        assert dag_mat is not None
        assert dag_mat.run_id == ""

        mapped_asset_mat = instance.get_latest_materialization_event(AssetKey("a"))
        assert mapped_asset_mat is not None
        assert mapped_asset_mat.run_id == ""


def test_monitor_sensor_cursor(init_load_context: None, instance: DagsterInstance) -> None:
    """Test that monitoring job correctly represents state in Dagster."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)

    with freeze_time(freeze_datetime):
        defs, af_instance = create_defs_and_instance(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            create_runs=False,
            create_assets_defs=False,
        )
        defs = build_job_based_airflow_defs(
            airflow_instance=af_instance,
            mapped_defs=defs,
        )
        context = build_sensor_context(
            instance=instance,
            repository_def=defs.get_repository_def(),
        )
        result = defs.sensors[0](context)  # type: ignore
        assert isinstance(result, RunRequest)
        assert result.run_config["ops"][monitoring_job_op_name(af_instance)] == {
            "config": {
                "range_start": (freeze_datetime - timedelta(days=1)).isoformat(),
                "range_end": freeze_datetime.isoformat(),
            }
        }
        assert result.tags["range_start"] == (freeze_datetime - timedelta(days=1)).isoformat()
        assert result.tags["range_end"] == freeze_datetime.isoformat()
        # Create an actual run for the monitoring job that is not finished.
        run = instance.create_run_for_job(
            job_def=defs.resolve_job_def(monitoring_job_name(af_instance.name)),
            run_id=make_new_run_id(),
            tags=result.tags,
            status=DagsterRunStatus.STARTED,
            run_config=result.run_config,
        )
        result = defs.sensors[0](context)  # type: ignore
        assert isinstance(result, SkipReason)
        assert "Monitoring job is still running" in result.skip_message  # type: ignore
        # Move the run to a finished state.
        instance.report_dagster_event(
            run_id=run.run_id,
            dagster_event=DagsterEvent(
                event_type_value="PIPELINE_SUCCESS",
                job_name=job_name(af_instance.name),
            ),
        )
    # Move time forward and check that we get a new run request.
    with freeze_time(freeze_datetime + timedelta(seconds=30)):
        result = defs.sensors[0](context)  # type: ignore
        assert isinstance(result, RunRequest)
        assert result.run_config["ops"][monitoring_job_op_name(af_instance)] == {
            "config": {
                "range_start": (freeze_datetime).isoformat(),
                "range_end": (freeze_datetime + timedelta(seconds=30)).isoformat(),
            }
        }


def test_metadata_from_step_output_events() -> None:
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)

    defs, af_instance = create_defs_and_instance(
        assets_per_task={
            "dag": {"task": [("a", [])]},
        },
        create_runs=False,
        create_assets_defs=False,
        seeded_runs=[
            # Have a newly started run.
            make_dag_run(
                dag_id="dag",
                run_id="run-dag",
                start_date=freeze_datetime - timedelta(seconds=30),
                end_date=None,
                state="running",
            ),
        ],
        seeded_task_instances=[
            # Have a newly completed task instance for the newly started run.
            make_task_instance(
                dag_id="dag",
                task_id="task",
                run_id="run-dag",
                start_date=freeze_datetime - timedelta(seconds=30),
                end_date=freeze_datetime,
            ),
        ],
    )
    defs = build_job_based_airflow_defs(
        airflow_instance=af_instance,
        mapped_defs=defs,
    )

    @asset
    def a(context: AssetExecutionContext):
        context.add_output_metadata({"foo": "bar"})

    with instance_for_test() as instance:
        # Simulate a run of the job with step output events.
        result = materialize(
            [a],
            tags={
                DAG_ID_TAG_KEY: "dag",
                DAG_RUN_ID_TAG_KEY: "run-dag",
                TASK_ID_TAG_KEY: "task",
                EXTERNALLY_MANAGED_ASSETS_TAG: "true",
            },
            instance=instance,
        )
        assert result.success
        assert result.get_asset_materialization_events() == []
        assert instance.get_latest_materialization_event(AssetKey("a")) is None

        # Now run the monitoring job. The resulting materialization should have metadata.
        result = definitions_execute_job_in_process(
            defs=defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            tags={REPOSITORY_LABEL_TAG: "placeholder"},
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(freeze_datetime - timedelta(seconds=30)).isoformat(),
                        range_end=freeze_datetime.isoformat(),
                    )
                }
            ),
        )

        run_record = next(
            iter(
                instance.get_run_records(
                    filters=RunsFilter(
                        tags={DAG_RUN_ID_TAG_KEY: "run-dag", OBSERVATION_RUN_TAG_KEY: "true"}
                    )
                )
            )
        )
        assert run_record.dagster_run.status == DagsterRunStatus.STARTED
        mat = instance.get_latest_materialization_event(AssetKey("a"))
        assert mat is not None
        assert mat.asset_materialization.metadata == {"foo": TextMetadataValue("bar")}  # type: ignore


def test_monitoring_sensor_run_fails(instance: DagsterInstance) -> None:
    """Ensure that sensor does not advance if the monitoring job run fails."""
    freeze_datetime = datetime(2021, 1, 1, tzinfo=timezone.utc)

    with freeze_time(freeze_datetime):
        defs, af_instance = create_defs_and_instance(
            assets_per_task={
                "dag": {"task": [("a", [])]},
            },
            create_runs=False,
            create_assets_defs=False,
        )
        defs = build_job_based_airflow_defs(
            airflow_instance=af_instance,
            mapped_defs=defs,
        )
        context = build_sensor_context(
            instance=instance,
            repository_def=defs.get_repository_def(),
        )
        result = defs.sensors[0](context)  # type: ignore
        assert isinstance(result, RunRequest)
        assert result.run_config["ops"][monitoring_job_op_name(af_instance)] == {
            "config": {
                "range_start": (freeze_datetime - timedelta(days=1)).isoformat(),
                "range_end": freeze_datetime.isoformat(),
            }
        }
        assert result.tags["range_start"] == (freeze_datetime - timedelta(days=1)).isoformat()
        assert result.tags["range_end"] == freeze_datetime.isoformat()
        # Create an actual run for the monitoring job that is failed.
        run = instance.create_run_for_job(
            job_def=defs.resolve_job_def(monitoring_job_name(af_instance.name)),
            run_id=make_new_run_id(),
            tags=result.tags,
            status=DagsterRunStatus.FAILURE,
            run_config=result.run_config,
        )
        with pytest.raises(Exception):
            defs.sensors[0](context)  # type: ignore

        # Create a successful run for the monitoring job within the same run group.
        instance.create_run_for_job(
            job_def=defs.resolve_job_def(monitoring_job_name(af_instance.name)),
            run_id=make_new_run_id(),
            status=DagsterRunStatus.SUCCESS,
            tags={ROOT_RUN_ID_TAG: run.run_id, **run.tags},
            run_config=result.run_config,
        )

    with freeze_time(freeze_datetime + timedelta(seconds=30)):
        # Now, the sensor should advance.
        result = defs.sensors[0](context)  # type: ignore
        assert isinstance(result, RunRequest)
        assert result.run_config["ops"][monitoring_job_op_name(af_instance)] == {
            "config": {
                "range_start": (freeze_datetime).isoformat(),
                "range_end": (freeze_datetime + timedelta(seconds=30)).isoformat(),
            }
        }



from datetime import datetime
from pathlib import Path

from airflow import DAG
from airflow.models import BaseOperator
from airflow.operators.python import PythonOperator
from dagster_airlift.in_airflow import DefaultProxyTaskToDagsterOperator, proxying_to_dagster
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml
from dagster_airlift.in_airflow.task_proxy_operator import (
    BaseProxyTaskToDagsterOperator,
    build_dagster_task,
)


def print_hello() -> None:
    print("Hello")  # noqa: T201


def build_print_task(task_id: str) -> PythonOperator:
    return PythonOperator(task_id=task_id, python_callable=print_hello, doc_md="Original doc")


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(2023, 1, 1),
    "retries": 0,
}


with DAG(
    "affected_dag",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as affected_dag:
    build_print_task("print_task") >> build_print_task("downstream_print_task")  # type: ignore

with DAG(
    "unaffected_dag",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
) as unaffected_dag:
    build_print_task("print_task") >> build_print_task("downstream_print_task")  # type: ignore


class CustomProxyTaskToDagsterOperator(DefaultProxyTaskToDagsterOperator):
    @classmethod
    def build_from_task(cls, original_task: BaseOperator) -> BaseProxyTaskToDagsterOperator:  # pyright: ignore[reportIncompatibleMethodOverride]
        """A custom callback to construct a new operator for the given task. In our case, we add retries to the affected_dag."""
        new_task = build_dagster_task(original_task, DefaultProxyTaskToDagsterOperator)
        if original_task.dag_id == "affected_dag":
            assert original_task.start_date
            new_task.retries = 1
        return new_task


proxying_to_dagster(
    proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
    global_vars=globals(),
    build_from_task_fn=CustomProxyTaskToDagsterOperator.build_from_task,
)



import argparse

from perf_harness.shared.constants import get_num_dags, get_num_tasks
from perf_harness.shared.utils import scaffold_proxied_state

# Set up argument parser
parser = argparse.ArgumentParser(description="Generate YAML files with a specified proxied state.")
parser.add_argument("proxied_state", type=str, help="The proxied state to use in the YAML files")

# Parse arguments
args = parser.parse_args()

scaffold_proxied_state(get_num_dags(), get_num_tasks(), args.proxied_state == "True")



import logging
from datetime import datetime

from airflow import DAG
from dagster_airlift.in_airflow.task_proxy_operator import DefaultProxyTaskToDagsterOperator

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)
requests_log = logging.getLogger("requests.packages.urllib3")
requests_log.setLevel(logging.INFO)
requests_log.propagate = True


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "retries": 1,
}

dag = DAG(
    "the_dag",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
    start_date=datetime(2023, 1, 1),
)
print_task = DefaultProxyTaskToDagsterOperator(task_id="some_task", dag=dag)
other_task = DefaultProxyTaskToDagsterOperator(task_id="other_task", dag=dag)



import logging
import os
from datetime import datetime

import requests
from airflow import DAG
from airflow.utils.context import Context
from dagster_airlift.in_airflow.materialize_assets_operator import BaseMaterializeAssetsOperator

logging.basicConfig()
logging.getLogger().setLevel(logging.INFO)
requests_log = logging.getLogger("requests.packages.urllib3")
requests_log.setLevel(logging.INFO)
requests_log.propagate = True


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "retries": 1,
}

dag = DAG(
    "the_dag",
    default_args=default_args,
    schedule_interval=None,
    is_paused_upon_creation=False,
    start_date=datetime(2023, 1, 1),
)


class BlankSessionAssetsOperator(BaseMaterializeAssetsOperator):
    """An assets operator which opens a blank session and expects the dagster URL to be set in the environment.
    The dagster url is expected to be set in the environment as DAGSTER_URL.
    """

    def get_dagster_session(self, context: Context) -> requests.Session:  # pyright: ignore[reportIncompatibleMethodOverride]
        return requests.Session()

    def get_dagster_url(self, context: Context) -> str:  # pyright: ignore[reportIncompatibleMethodOverride]
        return os.environ["DAGSTER_URL"]


the_task = BlankSessionAssetsOperator(
    # Test both string syntax and list of strings syntax.
    task_id="some_task",
    dag=dag,
    asset_key_paths=[
        "some_asset",
        ["other_asset"],
        ["nested", "asset"],
        "string/interpretation",
        r"backslash\/interpretation",
    ],
)



import requests
from dagster import AssetExecutionContext, asset
from dagster._core.storage.tags import (
    AUTO_RETRY_RUN_ID_TAG,
    MAX_RETRIES_TAG,
    PARENT_RUN_ID_TAG,
    RETRY_NUMBER_TAG,
    WILL_RETRY_TAG,
)
from dagster_airlift.in_airflow.base_asset_operator import build_dagster_run_execution_params
from dagster_airlift.in_airflow.gql_queries import TRIGGER_ASSETS_MUTATION


# Asset that simulates having run retries activated (so that we don't have to stand up non-sqlite-storage)
@asset
def succeeds_on_final_retry(context: AssetExecutionContext):
    if RETRY_NUMBER_TAG not in context.run_tags or int(context.run_tags[RETRY_NUMBER_TAG]) < 2:
        # Launch a run of the "next retry". We need to do it in a separate request;
        # launching the job within the same request will cause weird context behavior.
        current_retry = int(context.run_tags.get(RETRY_NUMBER_TAG, 0))
        response = requests.post(
            "http://localhost:3333/graphql",
            json={
                "query": TRIGGER_ASSETS_MUTATION,
                "variables": {
                    "executionParams": build_dagster_run_execution_params(
                        tags={
                            **context.run_tags,
                            RETRY_NUMBER_TAG: str(current_retry + 1),
                            PARENT_RUN_ID_TAG: context.run_id,
                            MAX_RETRIES_TAG: "3",
                        },
                        job_identifier=(
                            "kitchen_sink.dagster_defs.mapped_defs",
                            "__repository__",
                            "__ASSET_JOB",
                        ),
                        asset_key_paths=["succeeds_on_final_retry"],
                    )
                },
            },
            timeout=10,
        )
        context.log.info(f"Received response: {response.json()}")
        retry_run_id = response.json()["data"]["launchPipelineExecution"]["run"]["id"]
        # Add a tag (before the run fails) to indicate that the current run will be retried
        context.log.info(f"Adding tag {WILL_RETRY_TAG} to run {context.run_id}")
        context.instance.run_storage.add_run_tags(
            context.run_id, {WILL_RETRY_TAG: "true", AUTO_RETRY_RUN_ID_TAG: retry_run_id}
        )
        raise Exception("oops i failed")
    return None


@asset
def just_fails():
    raise Exception("I fail every time")



from datetime import datetime

import pytest
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.test import (
    AirflowInstanceFake,
    make_dag_info,
    make_dag_run,
    make_task_info,
    make_task_instance,
)


def test_test_instance() -> None:
    """Test the AirflowInstanceFake class."""
    dag_info = make_dag_info(
        instance_name="test", dag_props={}, dag_id="test_dag", file_token="test_file_token"
    )
    task_info = make_task_info(dag_id="test_dag", task_id="test_task")
    task_instance = make_task_instance(
        dag_id="test_dag",
        task_id="test_task",
        run_id="test_run_id",
        start_date=datetime(2022, 1, 1),
        end_date=datetime(2022, 1, 2),
    )
    assert task_instance.logical_date == datetime(2022, 1, 1)
    dag_run = make_dag_run(
        dag_id="test_dag",
        run_id="test_run_id",
        start_date=datetime(2022, 1, 1),
        end_date=datetime(2022, 1, 2),
    )
    assert dag_run.logical_date == datetime(2022, 1, 1)
    test_instance = AirflowInstanceFake(
        dag_infos=[dag_info],
        task_infos=[task_info],
        task_instances=[task_instance],
        dag_runs=[dag_run],
    )

    # Dag doesn't exist in instance
    with pytest.raises(ValueError):
        test_instance.get_dag_runs(
            dag_id="nonexistent_dag", start_date=datetime(2022, 1, 1), end_date=datetime(2022, 1, 2)
        )

    # Task doesn't exist in instance
    with pytest.raises(ValueError):
        test_instance.get_task_info(dag_id="test_dag", task_id="nonexistent_task")

    # Dag not in instance
    with pytest.raises(ValueError):
        test_instance.get_task_infos(dag_id="nonexistent_dag")

    # Task instance doesn't exist in instance (task id is wrong)
    with pytest.raises(ValueError):
        test_instance.get_task_instance(
            dag_id="test_dag", task_id="nonexistent_task", run_id="test_run_id"
        )

    # Task instance doesn't exist in instance (dag id is wrong)
    with pytest.raises(ValueError):
        test_instance.get_task_instance(
            dag_id="nonexistent_dag", task_id="test_task", run_id="test_run_id"
        )

    # Task instance doesn't exist in instance (run id is wrong)
    with pytest.raises(ValueError):
        test_instance.get_task_instance(
            dag_id="test_dag", task_id="test_task", run_id="nonexistent_run_id"
        )

    with pytest.raises(ValueError):
        test_instance.get_dag_runs(
            dag_id="nonexistent_dag", start_date=datetime(2022, 1, 1), end_date=datetime(2022, 1, 2)
        )

    assert test_instance.list_dags(retrieval_filter=AirflowFilter()) == [dag_info]
    assert test_instance.get_dag_info(dag_id="test_dag") == dag_info
    # Matching range
    assert test_instance.get_dag_runs(
        dag_id="test_dag", start_date=datetime(2022, 1, 1), end_date=datetime(2022, 1, 2)
    ) == [dag_run]
    # Range is too late
    assert (
        test_instance.get_dag_runs(
            dag_id="test_dag", start_date=datetime(2022, 1, 2), end_date=datetime(2022, 1, 3)
        )
        == []
    )
    # Range is too early
    assert (
        test_instance.get_dag_runs(
            dag_id="test_dag", start_date=datetime(2021, 12, 31), end_date=datetime(2022, 1, 1)
        )
        == []
    )

    # Can retrieve task info
    assert test_instance.get_task_info(dag_id="test_dag", task_id="test_task") == task_info

    # Can retrieve task infos
    assert test_instance.get_task_infos(dag_id="test_dag") == [task_info]

    # Can retrieve task instance
    assert (
        test_instance.get_task_instance(
            dag_id="test_dag", task_id="test_task", run_id="test_run_id"
        )
        == task_instance
    )

    assert (
        test_instance.get_dag_source_code(file_token="test_file_token")
        == "indicates found source code"
    )
    with pytest.raises(ValueError):
        test_instance.get_dag_source_code(file_token="nonexistent_file_token")



from pathlib import Path

import yaml

target_dir = Path(__file__).parent.parent / "airflow_dags" / "proxied_state"
# Ensure the target directory exists
target_dir.mkdir(parents=True, exist_ok=True)


def scaffold_proxied_state(num_dags: int, num_tasks: int, proxied_state: bool):
    # Ensure the target directory exists
    target_dir.mkdir(parents=True, exist_ok=True)

    for i in range(num_dags):
        yaml_dict = {
            "tasks": [{"id": f"task_{i}_{j}", "proxied": proxied_state} for j in range(num_tasks)],
        }
        # Write to a file dag_{i}.yaml
        with open(target_dir / f"dag_{i}.yaml", "w") as f:
            yaml.dump(yaml_dict, f)



from datetime import datetime, timedelta

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.metadata.metadata_value import TimestampMetadataValue
from dagster._core.definitions.run_config import RunConfig
from dagster._core.event_api import EventRecordsFilter
from dagster._core.events import DagsterEventType
from dagster._core.instance_for_test import instance_for_test
from dagster._core.storage.dagster_run import DagsterRunStatus
from dagster._core.storage.tags import EXTERNAL_JOB_SOURCE_TAG_KEY
from dagster._utils.test.definitions import scoped_definitions_load_context
from dagster_airlift.constants import DAG_ID_TAG_KEY, DAG_RUN_ID_TAG_KEY
from dagster_airlift.core.monitoring_job.builder import MonitoringConfig, monitoring_job_op_name
from dagster_airlift.core.utils import monitoring_job_name
from dagster_airlift.test.test_utils import asset_spec
from dagster_test.utils.definitions_execute_in_process import definitions_execute_job_in_process
from kitchen_sink.airflow_instance import local_airflow_instance

from kitchen_sink_tests.integration_tests.conftest import (
    poll_for_airflow_run_existence_and_completion,
)


def test_component_based_defs(
    airflow_instance: None,
) -> None:
    """Test that component based defs load properly."""
    with instance_for_test() as instance:
        with scoped_definitions_load_context():
            from kitchen_sink.dagster_defs.component_defs import defs

        assert len(defs.jobs) == 21  # type: ignore
        assert len(defs.assets) == 2  # type: ignore
        assert len(defs.sensors) == 1  # type: ignore
        for key in ["example1", "example2"]:
            assert asset_spec(key, defs)

        # First, execute dataset producer dag
        af_instance = local_airflow_instance("kitchen_sink_instance")
        af_run_id = af_instance.trigger_dag("dataset_producer")
        poll_for_airflow_run_existence_and_completion(
            af_instance=af_instance, af_run_id=af_run_id, dag_id="dataset_producer", duration=30
        )

        # Then, execute monitoring job
        result = definitions_execute_job_in_process(
            defs=defs,
            job_name=monitoring_job_name(af_instance.name),
            instance=instance,
            run_config=RunConfig(
                ops={
                    monitoring_job_op_name(af_instance): MonitoringConfig(
                        range_start=(datetime.now() - timedelta(seconds=30)).isoformat(),
                        range_end=datetime.now().isoformat(),
                    )
                }
            ),
        )
        assert result.success
        # There should be a run for the dataset producer dag
        runs = instance.get_runs()
        assert len(runs) == 2
        producer_run = next(run for run in runs if run.job_name == "dataset_producer")
        assert producer_run.status == DagsterRunStatus.SUCCESS
        assert producer_run.tags[DAG_RUN_ID_TAG_KEY] == af_run_id
        assert producer_run.tags[DAG_ID_TAG_KEY] == "dataset_producer"
        assert producer_run.tags[EXTERNAL_JOB_SOURCE_TAG_KEY] == "airflow"

        # Check that there are asset planned events for the two assets
        planned_records = instance.get_event_records(
            event_records_filter=EventRecordsFilter(
                event_type=DagsterEventType.ASSET_MATERIALIZATION_PLANNED
            ),
        )
        assert len(planned_records) == 2
        assert {planned_records.asset_key for planned_records in planned_records} == {
            AssetKey("example1"),
            AssetKey("example2"),
        }
        assert {planned_records.run_id for planned_records in planned_records} == {
            producer_run.run_id
        }

        # Check that there are asset materialized events for the two assets
        materialized_records = instance.get_event_records(
            event_records_filter=EventRecordsFilter(
                event_type=DagsterEventType.ASSET_MATERIALIZATION
            ),
        )
        assert len(materialized_records) == 2
        assert {
            materialized_records.asset_key for materialized_records in materialized_records
        } == {
            AssetKey("example1"),
            AssetKey("example2"),
        }
        assert {materialized_records.run_id for materialized_records in materialized_records} == {
            producer_run.run_id
        }
        for key in ["example1", "example2"]:
            key_record = next(
                materialized_records
                for materialized_records in materialized_records
                if materialized_records.asset_key == AssetKey(key)
            )
            metadata = key_record.asset_materialization.metadata  # type: ignore
            assert metadata["my_timestamp"] == TimestampMetadataValue(value=111.0)
            assert metadata["my_other_timestamp"] == TimestampMetadataValue(value=113.0)
            assert metadata["foo"].value == "baz"



# ruff: noqa: T201
import time

# Start timing for imports
import_start_time = time.time()
from pathlib import Path

from airflow import DAG
from airflow.operators.python import PythonOperator
from dagster._time import get_current_datetime
from dagster_airlift.in_airflow import proxying_to_dagster
from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml

from perf_harness.shared.constants import get_num_dags, get_num_tasks

# End timing for imports
import_end_time = time.time()
import_time = import_end_time - import_start_time

# Start timing for DAG creation
dag_creation_start_time = time.time()

default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": get_current_datetime(),
}
global_vars = globals()
for i in range(get_num_dags()):
    dag = DAG(
        dag_id=f"dag_{i}",
        default_args=default_args,
        is_paused_upon_creation=False,
    )
    for j in range(get_num_tasks()):
        global_vars[f"task_{i}_{j}"] = PythonOperator(
            python_callable=lambda: print(f"Task {i}_{j}"),
            task_id=f"task_{i}_{j}",
            dag=dag,
        )
    global_vars[f"dag_{i}"] = dag

# End timing for DAG creation
dag_creation_end_time = time.time()
dag_creation_time = dag_creation_end_time - dag_creation_start_time

# Start timing for proxying_to_dagster
proxying_to_dagster_start_time = time.time()

proxying_to_dagster(
    global_vars=globals(),
    proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
)

# End timing for proxying_to_dagster
mark_as_dagster_end_time = time.time()
mark_as_dagster_time = mark_as_dagster_end_time - proxying_to_dagster_start_time

# Calculate total time
total_time = mark_as_dagster_end_time - import_start_time



import os
import subprocess
from collections.abc import Generator
from pathlib import Path

import pytest
from dagster._core.test_utils import environ
from dagster_airlift.test.shared_fixtures import stand_up_airflow


def makefile_dir() -> Path:
    return Path(__file__).parent.parent


@pytest.fixture(name="local_env")
def local_env_fixture() -> Generator[None, None, None]:
    subprocess.run(["make", "setup_local_env"], cwd=makefile_dir(), check=True)
    with environ(
        {
            "AIRFLOW_HOME": str(makefile_dir() / ".airflow_home"),
            "DAGSTER_HOME": str(makefile_dir() / ".dagster_home"),
        }
    ):
        yield
    subprocess.run(["make", "wipe"], cwd=makefile_dir(), check=True)


@pytest.fixture(name="dags_dir")
def dags_dir_fixture() -> Path:
    return Path(__file__).parent.parent / "perf_harness" / "airflow_dags"


@pytest.fixture(name="airflow_home")
def airflow_home_fixture(local_env: None) -> Path:
    return Path(os.environ["AIRFLOW_HOME"])


@pytest.fixture(name="airflow_instance")
def airflow_instance_fixture(local_env: None) -> Generator[subprocess.Popen, None, None]:
    with stand_up_airflow(
        airflow_cmd=["make", "run_airflow"], env=os.environ, cwd=makefile_dir()
    ) as process:
        yield process



import_failed = False
try:
    import dagster as dagster
except ImportError:
    import_failed = True
if not import_failed:
    raise Exception("Importing dagster did not fail")

requires_airflow = False
try:
    from dagster_airlift.in_airflow import AirflowProxiedState as AirflowProxiedState
except Exception as e:
    if "Airflow is not installed" in str(e):
        requires_airflow = True
if not requires_airflow:
    raise Exception("Importing dagster_airlift.in_airflow did not fail")



import subprocess
from pathlib import Path

expected_lines = [
    "Total airflow dags load time",
    "Airflow defs import time",
    "Proxying to dagster time",
    "Airflow setup time",
    "Airflow standup time",
    "Peered defs import time",
    "Observed defs import time",
    "Migrate defs import time",
    "peer defs initial load time",
    "observe defs initial load time",
    "migrate defs initial load time",
    "peer sensor tick with no runs time",
    "observe sensor tick with no runs time",
    "migrate sensor tick with no runs time",
    "peer sensor tick with single run time",
    "observe sensor tick with single run time",
    "migrate sensor tick with single run time",
    "peer sensor tick with 1 runs time",
    "observe sensor tick with 1 runs time",
    "migrate sensor tick with 1 runs time",
]

makefile_dir = Path(__file__).parent.parent
expected_file = makefile_dir / "perf_harness" / "shared" / "1_dags_1_tasks_perf_output.txt"


def test_cli() -> None:
    """Test that the CLI can be run, and produces expected output."""
    subprocess.call(
        ["make", "run_perf_scenarios_test"], cwd=makefile_dir, stdout=subprocess.DEVNULL
    )
    assert expected_file.exists()
    for i, line in enumerate(expected_file.read_text().split("\n")[:-1]):  # last line is empty
        assert expected_lines[i] in line
